from fractions import Fraction

import pytest
from hypothesis import example, given, strategies

from measured import Length, Number, One, Unit
from measured.hypothesis import base_units, units, units_with_symbols
from measured.iec import Byte
from measured.si import Liter, Meter, Second


@pytest.fixture(scope="module")
def identity() -> Unit:
    return One


@given(a=units(), b=units())
def test_homogenous_under_addition(a: Unit, b: Unit) -> None:
    # https://en.wikipedia.org/wiki/Dimensional_analysis#Dimensional_homogeneity
    #
    # Only commensurable quantities (physical quantities having the same dimension) may
    # be compared, equated, added, or subtracted.
    if a is b:
        assert a + b == a
    else:
        with pytest.raises(TypeError):
            a + b


@given(a=units(), b=units())
def test_homogenous_under_subtraction(a: Unit, b: Unit) -> None:
    # https://en.wikipedia.org/wiki/Dimensional_analysis#Dimensional_homogeneity
    #
    # Only commensurable quantities (physical quantities having the same dimension) may
    # be compared, equated, added, or subtracted.
    if a is b:
        assert a - b == a
    else:
        with pytest.raises(TypeError):
            a - b


@given(a=units(), b=units(), c=units())
def test_abelian_associativity(a: Unit, b: Unit, c: Unit) -> None:
    # https://en.wikipedia.org/wiki/Abelian_group
    assert (a * b) * c == a * (b * c)


@given(a=units())
def test_abelian_identity(identity: Unit, a: Unit) -> None:
    assert identity * a == a


@given(a=units())
@example(a=Byte)
def test_abelian_inverse(identity: Unit, a: Unit) -> None:
    inverse = a**-1
    assert inverse * a == a * inverse
    assert inverse * a == identity
    assert identity / a == inverse


@given(a=units(), b=units())
def test_abelian_commutativity(a: Unit, b: Unit) -> None:
    assert a * b == b * a


@given(unit=units())
def test_no_dimensional_exponentation(unit: Unit) -> None:
    with pytest.raises(TypeError):
        unit**unit  # type: ignore


@given(unit=units())
def test_no_floating_point_exponentation(unit: Unit) -> None:
    with pytest.raises(TypeError):
        unit**0.5  # type: ignore


@given(unit=units())
def test_no_fractional_exponentation(unit: Unit) -> None:
    with pytest.raises(TypeError):
        unit ** Fraction(1, 2)  # type: ignore


@given(unit=units())
def test_repr_roundtrips(unit: Unit) -> None:
    from measured import Dimension, Prefix  # noqa: F401

    assert eval(repr(unit)) is unit


@given(base=base_units())
def test_base_units_factor_to_themselves(base: Unit) -> None:
    assert base.factors == {base: 1}


def test_one() -> None:
    assert isinstance(One, Unit)
    assert One.dimension is Number
    assert One.name == "one"
    assert One.symbol == "1"


@given(root=strategies.integers())
def test_identity_to_any_root(root: int) -> None:
    assert One.root(0) == One
    assert One.root(2) == One
    assert One.root(3) == One
    assert One.root(100) == One


@given(unit=units())
def test_zeroth_root_always_one(unit: Unit) -> None:
    assert unit.root(0) is One


def test_roots() -> None:
    assert (Meter**2).root(0) == One
    assert (Meter**2).root(2) == Meter
    assert (Meter**4).root(2) == Meter**2
    assert (Meter**16 / Second**12).root(4) == Meter**4 / Second**3


def test_only_integer_roots() -> None:
    with pytest.raises(TypeError):
        (Meter**2).root(0.5)  # type: ignore


def test_whole_power_roots_only() -> None:
    with pytest.raises(ValueError):
        (Meter**3).root(2)


def test_names_unique() -> None:
    with pytest.raises(ValueError, match="already defined"):
        Length.unit("meter", "totally unique")

    with pytest.raises(ValueError, match="already defined"):
        Unit.derive(Meter / Second, name="ohm", symbol="totally unique")


def test_symbols_unique() -> None:
    with pytest.raises(ValueError, match="already defined"):
        Length.unit("totally unique", "m")

    with pytest.raises(ValueError, match="already defined"):
        Unit.derive(Meter / Second, name="totally unique", symbol="Ω")


@given(unit=units_with_symbols())
@example(unit=Liter)  # liter is interesting in that it includes a prefix
def test_units_str_to_their_symbols(unit: Unit) -> None:
    assert str(unit) == unit.symbol



import pytest
from pytest import approx

from measured import (
    Bel,
    Decibel,
    Level,
    Logarithm,
    Neper,
    Numeric,
    Power,
    Pressure,
    Quantity,
    approximately,
)
from measured.acoustics import dBSPL
from measured.electronics import dBW
from measured.si import Deci, Meter, Micro, Milli, Volt, Watt


def test_logarithmic_units_are_singletons() -> None:
    assert Logarithm(base=10) is Bel
    assert (Deci * Bel) is Decibel


def test_logarithmic_units_are_not_compatible_with_linear_units() -> None:
    with pytest.raises(TypeError):
        assert Meter * Bel  # type: ignore


def test_logarithmic_units_can_be_multiplied_by_prefixes() -> None:
    decibel = Deci * Bel
    assert decibel is Decibel
    assert decibel.base == 10
    assert decibel.prefix == Deci

    millibel = Milli * Bel
    assert millibel.base == 10
    assert millibel.prefix == Milli

    microbel = Milli * (Milli * Bel)
    assert microbel.base == 10
    assert microbel.prefix == Micro


def test_logarithmic_units_do_not_produce_levels() -> None:
    with pytest.raises(TypeError):
        assert 10 * Bel  # type: ignore


def test_logarithmic_units_have_names_and_symbols() -> None:
    assert Bel.name == "bel"
    assert Bel.symbol == "bel"

    assert Decibel.name == "decibel"
    assert Decibel.symbol == "dB"

    assert Neper.name == "neper"
    assert Neper.symbol == "Np"


@pytest.mark.parametrize(
    "power, level",
    [
        (1 * Watt, 0 * dBW),
        (10 * Watt, 10 * dBW),
        (100 * Watt, 20 * dBW),
        (0.1 * Watt, -10 * dBW),
        (0.01 * Watt, -20 * dBW),
    ],
)
def test_quantity_to_level_comparisons(power: Quantity, level: Level) -> None:
    assert power == level
    assert level == power


@pytest.mark.parametrize(
    "power, level",
    [
        (1 * Watt, 0 * dBW),
        (10 * Watt, 10 * dBW),
        (100 * Watt, 20 * dBW),
        (0.1 * Watt, -10 * dBW),
        (0.01 * Watt, -20 * dBW),
    ],
)
def test_producing_level(power: Quantity, level: Level) -> None:
    assert power.level(dBW).unit is dBW
    assert power.level(dBW).magnitude == approx(level.magnitude)


@pytest.mark.parametrize(
    "power, level",
    [
        (1 * Watt, 0 * dBW),
        (10 * Watt, 10 * dBW),
        (100 * Watt, 20 * dBW),
        (0.1 * Watt, -10 * dBW),
        (0.01 * Watt, -20 * dBW),
    ],
)
def test_producing_quantity(power: Quantity, level: Level) -> None:
    assert level.quantify().unit is Watt
    assert level.quantify().magnitude == power.magnitude


@pytest.mark.parametrize(
    "bels, decibels",
    [
        (10, 100),
        (2, 20),
        (1, 10),
        (0.1, 1),
        (0.01, 0.1),
        (0, 0),
        (-1, -10),
        (-2, -20),
        (-10, -100),
    ],
)
def test_comparing_bels_to_decibels(bels: Numeric, decibels: Numeric) -> None:
    assert bels * Bel[1 * Watt] == decibels * dBW
    assert decibels * dBW == bels * Bel[1 * Watt]


@pytest.mark.parametrize(
    "decibels, nepers",
    [
        (10, 1.151277918),
        (8.685889638, 1),
        (2, 0.2302555836),
        (1, 0.1151277918),
        (0.1, 0.0115127792),
        (0.01, 0.0011512779),
        (0, 0),
        (-1, -0.1151277918),
        (-2, -0.2302555836),
        (-8.685889638, -1),
        (-10, -1.151277918),
    ],
)
def test_comparing_decibels_to_nepers(decibels: Numeric, nepers: Numeric) -> None:
    assert nepers * Neper[1 * Watt] == approximately(decibels * dBW, within=1e5)
    assert decibels * dBW == approximately(nepers * Neper[1 * Watt], within=1e5)


def test_logarithmic_addition() -> None:
    # Example adapted from
    # https://en.wikipedia.org/wiki/Decibel#Representation_of_addition_operations
    # But switched to the less controversial unit dBW for simplicity
    assert 70 * dBW + 90 * dBW == 90.04321373782642 * dBW


def test_addition_and_subtraction_only_by_levels() -> None:
    with pytest.raises(TypeError):
        (10 * dBW) + (1 * Watt)  # type: ignore

    with pytest.raises(TypeError):
        (10 * dBW) - (1 * Watt)  # type: ignore


def test_addition_and_subtraction_only_within_same_unit() -> None:
    with pytest.raises(TypeError):
        10 * dBW + 1 * Decibel[1 * Volt]

    with pytest.raises(TypeError):
        10 * dBW - 1 * Decibel[1 * Volt]


def test_logarithmic_subtraction() -> None:
    # Example adapted from
    # https://en.wikipedia.org/wiki/Decibel#Representation_of_addition_operations
    # But switched to the less controversial unit dBW for simplicity
    assert 87 * dBW - 83 * dBW == 84.79519169458092 * dBW


def test_logarithmic_multiplication() -> None:
    assert (10 * dBW) * 2 == 12 * dBW


def test_logarithmic_division() -> None:
    assert (10 * dBW) / 2 == 8 * dBW


def test_logarithmic_multiplication_only_by_numerics() -> None:
    with pytest.raises(TypeError):
        (10 * dBW) * (2 * dBW)  # type: ignore

    with pytest.raises(TypeError):
        (10 * dBW) / (2 * dBW)  # type: ignore


def test_power_ratios() -> None:
    # https://en.wikipedia.org/wiki/Power,_root-power,_and_field_quantities

    assert dBW.reference.unit.dimension is Power
    assert dBW.power_ratio == 1

    assert dBSPL.reference.unit.dimension is Pressure
    assert dBSPL.power_ratio == 2



# Units based on the meter, gram, and second, but not part of the SI system of units
# https://en.wikipedia.org/wiki/List_of_metric_units
#
# Includes specialized units compatible with SI and accepted for use with it by either
# the BIPM or NIST
# https://www.nist.gov/pml/special-publication-811/nist-guide-si-chapter-5-units-outside-si

from measured import Area, Frequency, Length, Time, Volume, VolumetricFlow
from measured.si import Femto, Hertz, Liter, Meter, Micro, Second, Tera

# Length, Area, and Volume

Ångström = Length.unit("Ångström", "Å")
Ångström.equals(1e-10 * Meter)

Fermi = Length.unit("fermi", "fermi")
Fermi.equals(1 * Femto * Meter)

Micron = Length.unit("micron", "μ")
Micron.equals(1 * Micro * Meter)

MetricMile = Length.unit("metric mile", "metric-mile")
MetricMile.equals(1500 * Meter)

Lieue = Length.unit("lieue", "lieue")
Lieue.equals(4000 * Meter)

Lambda = Volume.unit("lambda", "λ")
Lambda.equals(1 * Micro * Liter)

Stere = Volume.unit("stere", "st")
Stere.equals(1 * Meter**3)

# From the domain of high-energy and atomic physics
# https://en.wikipedia.org/wiki/Barn_(unit)
# https://en.wikipedia.org/wiki/Shake_(unit)

Barn = Area.unit("barn", "barn")
Barn.equals(1e-28 * Meter**2)

Outhouse = Area.unit("outhouse", "outhouse")
Outhouse.equals(1e-34 * Meter**2)

Shed = Area.unit("shed", "shed")
Shed.equals(1e-52 * Meter**2)


# Time

Svedberg = Time.unit("svedberg", "svedberg")
Svedberg.equals(100 * Femto * Second)

Shake = Time.unit("shake", "shake")
Shake.equals(1e-8 * Second)

Fresnel = Frequency.unit("fresnel", "fresnel")
Fresnel.equals(1 * Tera * Hertz)


# Derived

Sverdrup = VolumetricFlow.unit("sverdrup", "sverdrup")
Sverdrup.equals(1000000 * Meter**3 / Second)



import pytest

from measured import Length, Mass, Quantity, Temperature, Time, approximately
from measured.natural import PlanckLength, PlanckMass, PlanckTemperature, PlanckTime
from measured.physics import G, c, k, ℏ
from measured.si import Joule, Kelvin, Kilogram, Meter, Second


@pytest.mark.parametrize(
    "fundamental_constant, unity",
    [
        (c, 1 * (PlanckLength / PlanckTime)),
        (ℏ, 1 * ((PlanckLength**2 * PlanckMass) / PlanckTime)),
        (G, 1 * (PlanckLength**3 / (PlanckMass * PlanckTime**2))),
        (
            k,
            1
            * (
                (PlanckLength**2 * PlanckMass) / (PlanckTemperature * PlanckTime**2)
            ),
        ),
    ],
)
def test_unity(fundamental_constant: Quantity, unity: Quantity) -> None:
    assert fundamental_constant == approximately(unity)
    assert unity == approximately(fundamental_constant)


def test_hbar() -> None:
    assert ℏ == approximately(1.054571817e-34 * Joule * Second)


def test_length() -> None:
    assert PlanckLength.dimension is Length
    assert 1 * PlanckLength == approximately(1.616255e-35 * Meter, 1.3e-7)


def test_mass() -> None:
    assert PlanckMass.dimension is Mass
    assert 1 * PlanckMass == approximately(2.176434e-8 * Kilogram, 1.6e-7)


def test_time() -> None:
    assert PlanckTime.dimension is Time
    assert 1 * PlanckTime == approximately(5.391247e-44 * Second, 1.1e-7)


def test_temperature() -> None:
    assert PlanckTemperature.dimension is Temperature
    assert 1 * PlanckTemperature == approximately(1.416784e32 * Kelvin, 1.2e-7)



"""
Defines the [Avoirdupois system of weights][1] and their conversions to SI

[1]: https://en.wikipedia.org/wiki/Avoirdupois_system

Attributes: Base units

    Pound (Unit): The base unit for the Avoirdupois system

Attributes: Units of mass ("weight")

    Grain (Unit): 1/7000 lb.

    Dram (Unit): 1/256 lb.

    Ounce (Unit): 1/16 lb.

    LongHundredweight (Unit): 112 lb.

    LongTon (Unit): 2240 lb.

"""

from measured import Mass
from measured.si import Gram, Kilo, Milli

Grain = Mass.unit("grain", "gr.")
Grain.equals(64.79891 * Milli * Gram)

Dram = Mass.unit("dram", "dr.")
Dram.equals((27 + (11 / 32)) * Grain)
Dram.equals(1.7718451953125 * Gram)

Ounce = Mass.unit("ounce", "oz")
Ounce.equals(16 * Dram)
Ounce.equals(28.349523125 * Gram)

Pound = Mass.unit("pound", "lb.")
Pound.equals(16 * Ounce)
Pound.equals(453.59237 * Gram)

LongHundredweight = Mass.unit("long hundredweight", "long-hundredweight")
LongHundredweight.equals(112 * Pound)
LongHundredweight.equals(50.80234544 * Kilo * Gram)

LongTon = Mass.unit("long ton", "long-ton")
LongTon.equals(20 * LongHundredweight)
LongTon.equals(2240 * Pound)
LongTon.equals(1016.0469088 * Kilo * Gram)



import pytest

from measured import Length
from measured.si import Meter


def test_division_of_numbers_by_integer_quantities() -> None:
    assert 1 / 5 * Meter == 0.20 * Meter
    assert 1 / (5 * Meter) == 0.20 * Meter

    assert 4 / 2 * Meter == 2 * Meter
    assert 4 / (2 * Meter) == 2 * Meter


def test_division_of_numbers_by_float_quantities() -> None:
    assert 1.0 / 5 * Meter == 0.20 * Meter
    assert 1.0 / (5 * Meter) == 0.20 * Meter


def test_only_numbers_may_be_divided_by_quantities() -> None:
    with pytest.raises(TypeError):
        "wat" / (1 * Meter)  # type: ignore

    with pytest.raises(TypeError):
        Length / (1 * Meter)  # type: ignore



import operator
from functools import reduce
from typing import Any, Optional

from measured import Numeric, One, Quantity, Unit

from . import _parser
from .formatting import from_superscript

ParseError = _parser.LarkError


class QuantityTransformer(_parser.Transformer[Any, "Quantity"]):
    inline = _parser.v_args(inline=True)

    @inline
    def unit(self, numerator: Unit, denominator: Optional[Unit] = None) -> Unit:
        return numerator / (denominator or One)

    @inline
    def unit_sequence(self, *terms: Unit) -> Unit:
        return reduce(operator.mul, terms)

    @inline
    def term(self, symbol: str, exponent: int = 1) -> Unit:
        return Unit.resolve_symbol(symbol) ** exponent

    @inline
    def carat_exponent(self, exponent: str) -> int:
        return int(exponent[1:])

    @inline
    def superscript_exponent(self, exponent: str) -> int:
        value = from_superscript(exponent)
        assert isinstance(value, int)
        return value

    @inline
    def quantity(self, magnitude: Numeric, unit: Unit) -> "Quantity":
        return Quantity(magnitude, unit)

    int = inline(int)
    float = inline(float)


parser: _parser.Lark = _parser.Parser(transformer=QuantityTransformer())  # type: ignore



import pytest

from measured import Charge, Length, Mass, Quantity, Time, approximately
from measured.natural import AtomicCharge, AtomicLength, AtomicMass, AtomicTime
from measured.physics import c, e, kₑ, mₑ, α, ℏ
from measured.si import Coulomb, Kilogram, Meter, Second


@pytest.mark.parametrize(
    "fundamental_constant, unity",
    [
        (e, 1 * AtomicCharge),
        (mₑ, 1 * AtomicMass),
        (ℏ, 1 * ((AtomicLength**2 * AtomicMass) / AtomicTime)),
        (
            kₑ,
            (
                1
                * (AtomicLength**3 * AtomicMass)
                / (AtomicTime**2 * AtomicCharge**2)
            ),
        ),
        (c, α**-1 * (AtomicLength / AtomicTime)),
    ],
)
def test_unity(fundamental_constant: Quantity, unity: Quantity) -> None:
    assert fundamental_constant == approximately(unity)
    assert unity == approximately(fundamental_constant)


def test_length() -> None:
    assert AtomicLength.dimension is Length
    assert 1 * AtomicLength == approximately(5.292e-11 * Meter, within=5e-5)


def test_mass() -> None:
    assert AtomicMass.dimension is Mass
    assert 1 * AtomicMass == approximately(9.109e-31 * Kilogram, within=5e-5)


def test_time() -> None:
    assert AtomicTime.dimension is Time
    assert 1 * AtomicTime == approximately(2.419e-17 * Second, within=5e-5)


def test_charge() -> None:
    assert AtomicCharge.dimension is Charge
    assert 1 * AtomicCharge == approximately(1.6022e-19 * Coulomb, within=2e-5)



import json
from typing import Generator, Optional

import pytest
from pydantic import BaseModel

import measured.json
from measured import Dimension, Length, Prefix, Quantity, Unit
from measured.si import Kilo, Meter


@pytest.fixture
def codecs_installed() -> Generator[None, None, None]:
    with measured.json.codecs_installed():
        yield


class ExampleModel(BaseModel):
    dimension: Dimension = Length
    optional_dimension: Optional[Dimension] = None

    prefix: Prefix = Kilo
    optional_prefix: Optional[Prefix] = None

    unit: Unit = Meter
    optional_unit: Optional[Unit] = None

    quantity: Quantity = 5 * Meter
    optional_quantity: Optional[Quantity] = None


def test_dimension_field() -> None:
    model = ExampleModel(dimension=Length)
    assert model.dimension == Length
    assert model.dimension is Length


def test_prefix_field() -> None:
    model = ExampleModel(prefix=Kilo)
    assert model.prefix == Kilo
    assert model.prefix is Kilo


def test_unit_field() -> None:
    model = ExampleModel(unit=Meter)
    assert model.unit == Meter
    assert model.unit is Meter


def test_quantity_field() -> None:
    model = ExampleModel(quantity=2 * Meter)
    assert model.quantity == 2 * Meter


def test_dimension_field_from_string() -> None:
    model = ExampleModel(dimension="length")
    assert model.dimension == Length
    assert model.dimension is Length


def test_prefix_field_from_string() -> None:
    model = ExampleModel(prefix="kilo")
    assert model.prefix == Kilo
    assert model.prefix is Kilo


def test_unit_field_from_string() -> None:
    model = ExampleModel(unit="meter")
    assert model.unit == Meter
    assert model.unit is Meter


def test_dimension_field_from_none() -> None:
    model = ExampleModel(optional_dimension=None)
    assert model.optional_prefix is None


def test_prefix_field_from_none() -> None:
    model = ExampleModel(optional_prefix=None)
    assert model.optional_prefix is None


def test_unit_field_from_none() -> None:
    model = ExampleModel(optional_unit=None)
    assert model.optional_unit is None


def test_quantity_field_from_none() -> None:
    model = ExampleModel(optional_quantity=None)
    assert model.optional_quantity is None


def test_dimension_field_from_incompatible() -> None:
    with pytest.raises(ValueError):
        ExampleModel(dimension=11)


def test_prefix_field_from_incompatible() -> None:
    with pytest.raises(ValueError):
        ExampleModel(prefix=11)


def test_unit_field_from_incompatible() -> None:
    with pytest.raises(ValueError):
        ExampleModel(unit=11)


def test_quantity_field_from_incompatible() -> None:
    with pytest.raises(ValueError):
        ExampleModel(quantity=11)


def test_dimension_field_from_string_must_exist() -> None:
    with pytest.raises(ValueError, match="'flibbity' is not a named Dimension"):
        ExampleModel(dimension="flibbity")


def test_prefix_field_from_string_must_exist() -> None:
    with pytest.raises(ValueError, match="'kibbity' is not a named Prefix"):
        ExampleModel(prefix="kibbity")


def test_unit_field_from_string_must_exist() -> None:
    with pytest.raises(ValueError, match="'kibbity' is not a named Unit"):
        ExampleModel(unit="kibbity")


@pytest.fixture
def instance() -> ExampleModel:
    return ExampleModel()


def test_to_dict(instance: ExampleModel) -> None:
    assert instance.dict() == {
        "dimension": Length,
        "optional_dimension": None,
        "prefix": Kilo,
        "optional_prefix": None,
        "unit": Meter,
        "optional_unit": None,
        "quantity": 5 * Meter,
        "optional_quantity": None,
    }


def test_to_json_without_codecs_installed(instance: ExampleModel) -> None:
    with pytest.raises(TypeError, match="is not JSON serializable"):
        assert instance.json()


def test_to_json(codecs_installed: None, instance: ExampleModel) -> None:
    assert json.loads(instance.json(), cls=json.JSONDecoder, object_hook=None) == {
        "dimension": {
            "__measured__": "Dimension",
            "name": "length",
            "symbol": "L",
            "exponents": [0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
        },
        "optional_dimension": None,
        "prefix": {
            "__measured__": "Prefix",
            "name": "kilo",
            "symbol": "k",
            "base": 10,
            "exponent": 3,
        },
        "optional_prefix": None,
        "unit": {
            "__measured__": "Unit",
            "name": "meter",
            "symbol": "m",
            "prefix": None,
            "factors": None,
            "dimension": {
                "__measured__": "Dimension",
                "name": "length",
                "symbol": "L",
                "exponents": [0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
            },
        },
        "optional_unit": None,
        "quantity": {
            "__measured__": "Quantity",
            "magnitude": 5,
            "unit": {
                "__measured__": "Unit",
                "name": "meter",
                "symbol": "m",
                "prefix": None,
                "factors": None,
                "dimension": {
                    "__measured__": "Dimension",
                    "name": "length",
                    "symbol": "L",
                    "exponents": [0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
                },
            },
        },
        "optional_quantity": None,
    }



"""
Units and non-standardized conventions relating to acoustics.

Attributes: LogarithmicUnits

    dBSPL (LogarithmicUnit): Decibels of Sound `Pressure`, compared to 20 μPa

    dBSWL (LogarithmicUnit): Decibels of Sound `Power`, compared to 1 pW

    dBSIL (LogarithmicUnit): Decibels of Sound intensity, compared to 1 pW/m²

Attributes: Constants

    ABSOLUTE_THRESHOLD_OF_HEARING (Quantity, Pressure): 20 μPa, an approximation of the
        [minimum sound pressure][1] that can be heard by an average human ear

        [1]: https://en.wikipedia.org/wiki/Absolute_threshold_of_hearing
"""

from measured import Decibel
from measured.si import Meter, Micro, Pascal, Pico, Watt

# https://en.wikipedia.org/wiki/Absolute_threshold_of_hearing
ABSOLUTE_THRESHOLD_OF_HEARING = 20 * Micro * Pascal

# https://en.wikipedia.org/wiki/Sound_pressure#Sound_pressure_level
dBSPL = Decibel[ABSOLUTE_THRESHOLD_OF_HEARING].alias("dBSPL", "dBSPL")

# https://en.wikipedia.org/wiki/Sound_power#Sound_power_level
dBSWL = Decibel[1 * Pico * Watt].alias("dBSWL", "dBSWL")

# https://en.wikipedia.org/wiki/Sound_intensity#Sound_intensity_level
dBSIL = Decibel[1 * (Pico * Watt) / Meter**2].alias("dBSIL", "dBSIL")



"""
The goal of the `measured` library is to provide a sound foundation for recording and
converting physical quantities, while maintaining the integrity of their units and
dimensions.

While it aims to be the fastest library of its kind, automatically tracking the units
and dimensions of quantities introduces significant overhead.  You can use `measured`
for applications where the accuracy of the units is more important than raw numerical
computing speed.

The value classes of `measured` should generally be treated as immutable.  Rather than
setting the `magnitude` of a [`Quantity`][measured.Quantity], consider instantiating a
new one, or performing arithmetic on an existing one to produce new values.  Instances
of [`Dimension`][measured.Dimension], [`Prefix`][measured.Prefix], and
[`Unit`][measured.Unit] are always singletons within a Python process.

For more background on the approach taken by this library, please see the material on
[quantity calculus][1] and [dimensional analysis][2].  `measured` uses the terms
"dimension", "unit", "prefix", and "quantity" in the same way.


[1]: https://en.wikipedia.org/wiki/Quantity_calculus

[2]: https://en.wikipedia.org/wiki/Dimensional_analysis


Examples:

    >>> from measured.si import Meter, Second
    >>> distance = 10 * Meter
    >>> time = 2 * Second
    >>> speed = distance / time
    >>> assert speed == 5 * Meter / Second

Attributes: Fundamental dimensions

    Number (Dimension): the basis of [counting and measuring][1], used to define
        [dimensionless quantities][2]

        [1]: https://en.wikipedia.org/wiki/Number

        [2]: https://en.wikipedia.org/wiki/Dimensionless_quantity

    Length (Dimension): [distance][1] or extent through space

        [1]: https://en.wikipedia.org/wiki/Length

    Time (Dimension): [what clocks read][1], the intervals between events

        [1]: https://en.wikipedia.org/wiki/Time

    Mass (Dimension): how much [matter][1] is in a physical body

        [1]: https://en.wikipedia.org/wiki/Mass

    Charge (Dimension): how strongly matter [interacts with the electric field][1]

        [1]: https://en.wikipedia.org/wiki/Electric_charge

    Temperature (Dimension): the average amount of [kinetic energy][1] in a system

        [1]: https://en.wikipedia.org/wiki/Temperature

    AmountOfSubstance (Dimension): how many [elementary entities][1] are in an object

        [1]: https://en.wikipedia.org/wiki/Mole_(unit)

    LuminousIntensity (Dimension): how strong a [light source][1] is over a volume

        [1]: https://en.wikipedia.org/wiki/Luminous_intensity

    Information (Dimension): how much [entropy][1] is present in a random variable

        [1]: https://en.wikipedia.org/wiki/Information#Information_theory

Attributes: Derived dimensions

    Area (Dimension):

    Volume (Dimension):

    PlaneAngle (Dimension):

    SolidAngle (Dimension):

    Speed (Dimension):

    Acceleration (Dimension):

    Jerk (Dimension):

    Snap (Dimension):

    Crackle (Dimension):

    Pop (Dimension):

    Frequency (Dimension):

    Force (Dimension):

    Energy (Dimension):

    Power (Dimension):

    Current (Dimension):

    Potential (Dimension):

    Capacitance (Dimension):

    Resistance (Dimension):

    Conductance (Dimension):

    Inductance (Dimension):

    MagneticFlux (Dimension):

    MagneticBField (Dimension):

    LuminousFlux (Dimension):

    Illuminance (Dimension):

    RadioactiveDose (Dimension):

    Catalysis (Dimension):



Attributes: Base Prefixes

    IdentityPrefix (Prefix): represents the number 1, expressed as the prefix 0⁰


Attributes: Base Units

    One (Unit): represents the number 1, expressed as a unit of dimension `Number`

Attributes: Logarithms

    Bel (Logarithm): A logarithmic ratio in base 10
    Decibel (Logarithm): 1/10th of a Bel
    Neper (Logarithm): A logarithmic ratio in base _e_

"""

import math
from collections import defaultdict
from functools import lru_cache, total_ordering
from importlib.metadata import version
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generator,
    Iterable,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
    overload,
)

from . import formatting

try:
    from icecream import ic as _ic
except ImportError:  # pragma: no cover
    _ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa

ic = _ic

__version__ = version("measured")

NUMERIC_CLASSES = (int, float)
Numeric = Union[int, float]


class FractionalDimensionError(ValueError):
    """Raised when computing the nth root of a Dimension, Prefix, or Unit would result
    in a non-integer dimesion"""

    def __init__(self, degree: int, value: Union["Dimension", "Prefix", "Unit"]):
        super().__init__(
            f"Taking the root of degree {degree} of {value} would "
            "result in a fractional exponent"
        )


class Dimension:
    """Dimension represents the kind of physical quantity being measured.

    Unless you're doing something really cool, you probably won't instantiate new
    dimensions directly.  Instead, you'll import the base dimensions and combine them
    through multiplication, division, or exponentation.

    Attributes:
        name (Optional[str]): The name of this dimension, which may not be set in the
            case of complex dimensions

        symbol (Optional[str]): The conventional [dimensional
            analysis](https://en.wikipedia.org/wiki/Dimensional_analysis#Definition)
            symbol of this dimension

        exponents (Tuple[int, ...]): The exponents that define this dimension in terms
            of the fundamental dimensions

    Examples:

        >>> from measured import (
        ...     Number,
        ...     Length,
        ...     Time,
        ...     Mass,
        ...     Current,
        ...     Temperature,
        ...     AmountOfSubstance,
        ...     LuminousIntensity,
        ...     Information
        ... )
        >>> Length.symbol
        'L'
        >>> Length.name
        'length'

        These dimensions form the basis for more complex dimensions, which can be
        produced through multiplication, exponentation, and division.

        >>> from measured import Area, Volume, Frequency, Speed
        >>> assert Area == Length * Length
        >>> assert Volume == Length**3
        >>> assert Speed == Length / Time
        >>> assert Frequency == Number / Time

        `measured` maintains `Dimension` instances as singletons within a single
        process, so they can be used in both equality and identity tests.

        >>> assert Volume is Length**3

        Dimensions are hashable and may be used as keys in dictionaries.

        >>> lookup = {Volume: 'spacious'}
        >>> lookup[Volume]
        'spacious'

        Dimensions can be serialized in a number of ways, preserving their identity.

        >>> import pickle
        >>> assert pickle.loads(pickle.dumps(Volume)) is Volume

        >>> import json
        >>> from measured.json import MeasuredJSONEncoder, MeasuredJSONDecoder
        >>> json.dumps(Length, cls=MeasuredJSONEncoder)
        '{"__measured__": "Dimension", "name": "length", "symbol": "L", ...}'

        While using `measured`'s [JSON codecs](../serialization), Dimensions may be
        deserialized directly from that JSON representation.

        >>> encoded = json.dumps(Length, cls=MeasuredJSONEncoder)
        >>> json.loads(encoded, cls=MeasuredJSONDecoder)
        Dimension(exponents=(0, 1, 0, 0, 0, 0, 0, 0, 0, 0), name='length', symbol='L')

        With measured's JSON codecs installed, you can omit passing the encoder and
        decoder.

        >>> from measured.json import codecs_installed
        >>> with codecs_installed():
        ...     assert json.loads(json.dumps(Volume)) is Volume
    """

    _known: ClassVar[Dict[Tuple[int, ...], "Dimension"]] = {}
    _initialized: bool

    _fundamental: ClassVar[List["Dimension"]] = []
    _by_name: ClassVar[Dict[str, "Dimension"]] = {}

    __slots__ = ("_initialized", "exponents", "name", "symbol")

    exponents: Tuple[int, ...]
    name: Optional[str]
    symbol: Optional[str]

    def __new__(
        cls,
        exponents: Tuple[int, ...],
        name: Optional[str] = None,
        symbol: Optional[str] = None,
    ) -> "Dimension":
        key = exponents
        if key in cls._known:
            return cls._known[key]

        self = super().__new__(cls)
        self._initialized = False
        cls._known[key] = self
        return self

    def __init__(
        self,
        exponents: Tuple[int, ...],
        name: Optional[str] = None,
        symbol: Optional[str] = None,
    ) -> None:
        if self._initialized:
            return

        self.exponents = exponents
        self.name = name
        self.symbol = symbol
        self._initialized = True

        if name:
            self._by_name[name] = self

    @classmethod
    def fundamental(cls) -> Iterable["Dimension"]:
        """Returns the registered fundamental dimensions

        Fundamental dimensions are not derived from other dimensions, and they will
        always have an `exponents` tuple with a single `1`, like `[0, 0, 1, 0, 0...]`
        (except for `Number`, which is all `0`s).

        The fundamental dimensions are those defined directly using
        [`Dimension.define`][measured.Dimension.define], like those provided by
        `measured`.  The length and order of these Dimensions is also the length and
        order of each Dimensions `exponents` tuple.
        """
        return list(cls._fundamental)

    @classmethod
    def define(cls, name: str, symbol: str) -> "Dimension":
        """Defines a new fundamental Dimension

        Used by `measured` itself to define the fundamental Dimensions.  You may define
        additional dimensions, but be aware that doing so will change the cardinality of
        the `exponents` tuple for _all_ defined Dimensions.
        """
        index = len(cls._fundamental)
        if index == 0:
            # the first dimension must be Number, with an exponent of zero (identity)
            exponents = tuple([0])
        else:
            # other dimensions start with their exponent at one, and all others zero
            exponents = tuple(([0] * index) + [1])

        dimension = cls(exponents, name=name, symbol=symbol)

        # resize the keys for all previously known dimensions to account
        # for this new fundamental dimension
        for previous in list(cls._known.values()):
            del cls._known[previous.exponents]
            previous.exponents += (0,)
            cls._known[previous.exponents] = previous

        cls._fundamental.append(dimension)

        return dimension

    @classmethod
    def derive(
        cls, dimension: "Dimension", name: str, symbol: Optional[str] = None
    ) -> "Dimension":
        """Registers a new named dimension derived from other dimension"""
        dimension.name = name
        dimension.symbol = symbol or str(dimension)
        cls._by_name[name] = dimension
        return dimension

    def unit(self, name: str, symbol: str) -> "Unit":
        """Define a new unit of this dimension"""
        return Unit.define(self, name, symbol)

    def scale(self, zero: "Quantity", name: str, symbol: str) -> "Unit":
        """Define a new scale of this dimension, setting a zero point of another unit"""
        unit = self.unit(name, symbol)
        conversions.translate(unit, zero)
        return unit

    # Pickle support

    def __getnewargs_ex__(self) -> Tuple[Tuple[Tuple[int, ...]], Dict[str, Any]]:
        return (self.exponents,), {}

    # JSON support

    def __json__(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "symbol": self.symbol,
            "exponents": list(self.exponents),
        }

    @classmethod
    def __from_json__(cls, json_object: Dict[str, Any]) -> "Dimension":
        return Dimension(tuple(json_object["exponents"]))

    # Pydantic support

    @classmethod
    def __get_validators__(
        cls,
    ) -> Generator[Callable[[Union[str, "Dimension"]], "Dimension"], None, None]:
        yield cls.validate

    @classmethod
    def validate(cls, value: Union[str, "Dimension"]) -> "Dimension":
        if isinstance(value, str):
            if value not in cls._by_name:
                raise ValueError(f"{value!r} is not a named Dimension")

            return cls._by_name[value]

        if isinstance(value, Dimension):
            return value

        raise ValueError(f"No conversion from {value!r} to Dimension")

    __repr__ = formatting.dimension_repr
    __str__ = formatting.dimension_str
    _repr_pretty_ = formatting.dimension_pretty
    _repr_html_ = formatting.mathml(formatting.dimension_mathml)

    # https://en.wikipedia.org/wiki/Dimensional_analysis#Dimensional_homogeneity
    #
    # Only commensurable quantities (physical quantities having the same dimension)
    # may be compared, equated, added, or subtracted.

    def __add__(self, other: "Dimension") -> "Dimension":
        if self is not other:
            return NotImplemented

        return self

    def __sub__(self, other: "Dimension") -> "Dimension":
        if self is not other:
            return NotImplemented

        return self

    # https://en.wikipedia.org/wiki/Dimensional_analysis#Dimensional_homogeneity
    #
    # The dimensions form an [abelian
    # group](https://en.wikipedia.org/wiki/Abelian_group) under multiplication, so:
    # One may take ratios of incommensurable quantities (quantities with different
    # dimensions), and multiply or divide them.

    @staticmethod
    @lru_cache(maxsize=None)
    def _multiply(self: "Dimension", other: "Dimension") -> "Dimension":
        return Dimension(tuple(s + o for s, o in zip(self.exponents, other.exponents)))

    def __mul__(self, other: "Dimension") -> "Dimension":
        if not isinstance(other, Dimension):
            return NotImplemented

        return Dimension._multiply(self, other)

    @staticmethod
    @lru_cache(maxsize=None)
    def _divide(self: "Dimension", other: "Dimension") -> "Dimension":
        return Dimension(tuple(s - o for s, o in zip(self.exponents, other.exponents)))

    def __truediv__(self, other: "Dimension") -> "Dimension":
        if not isinstance(other, Dimension):
            return NotImplemented

        return Dimension._divide(self, other)

    def __pow__(self, power: int) -> "Dimension":
        if not isinstance(power, int):
            return NotImplemented

        return Dimension(tuple(s * power for s in self.exponents))

    def root(self, degree: int) -> "Dimension":
        """Returns the nth root of this Dimension"""
        if not isinstance(degree, int):
            raise TypeError(f"degree should be an integer, not {type(degree)}")

        if degree == 0:
            return Number

        if any(s // degree != s / degree for s in self.exponents):
            raise FractionalDimensionError(degree, self)

        return Dimension(tuple(s // degree for s in self.exponents))

    @lru_cache(maxsize=None)
    def as_ratio(self) -> Tuple["Dimension", "Dimension"]:
        """Returns this dimension, split into a numerator and denominator"""
        numerator = tuple(e if e >= 0 else 0 for e in self.exponents)
        denominator = tuple(-e if e < 0 else 0 for e in self.exponents)
        return Dimension(numerator), Dimension(denominator)

    def is_factor(self, other: "Dimension") -> bool:
        """Returns true if this dimension is a factor of the other dimension"""
        if self is other or self is Number:
            return True

        exponents = [
            i
            for i, (mine, theirs) in enumerate(zip(self.exponents, other.exponents))
            if (theirs and mine) and theirs >= mine
        ]
        return bool(exponents)


class Prefix:
    """Prefixes scale a [`Unit`][measured.Unit] up or down by a constant factor.

    Prefixes are defined in systems of factors with a common integer base, and
    successive positive and/or negative exponents, like the SI system of prefixes using
    base 10 (micro, milli, kilo, mega, etc) or the IEC system of binary prefixes (kibi,
    mebi, gibi, etc).

    Attributes:

        base (int): The base of the system, like `10` for the SI system
        exponent (int | float): The exponent that the base is raised to

    Examples:

        Prefixes have a base and an exponent:

        >>> from measured.si import Milli, Kilo
        >>> Milli.symbol, Milli.base, Milli.exponent
        ('m', 10, -3)
        >>> Kilo.symbol, Kilo.base, Kilo.exponent
        ('k', 10, 3)

        Prefixes can be multiplied by numbers, Prefixes, and Units:

        >>> from measured.si import Kilo, Mega, Meter
        >>> assert 1 * Kilo * Meter == 1000 * Meter
        >>> assert 1 * (Kilo * Kilo * Meter) == 1 * Mega * Meter

        Prefixes can be divided by Prefixes and other prefixed units:

        >>> from measured.si import Kilo, Mega, Meter, Second
        >>> assert (1 * Mega * Meter) / (1 * Kilo * Second) == 1000 * Meter / Second

        Prefixes can be raised to an exponent:

        >>> from measured.si import Kilo, Mega, Meter
        >>> assert (2 * Kilo * Meter)**2 == 4 * Mega * Meter**2

        Prefixes can translate between bases, although be careful with the loss of
        precision during this conversion:

        >>> from measured.si import Kilo
        >>> from measured.iec import Mebi, Bit
        >>> assert 1 * Mebi * Bit == 1048.576 * Kilo * Bit

    """

    _known: ClassVar[Dict[Tuple[int, Numeric], "Prefix"]] = {}
    _initialized: bool

    _by_name: ClassVar[Dict[str, "Prefix"]] = {}
    _by_symbol: ClassVar[Dict[str, "Prefix"]] = {}

    __slots__ = ("_initialized", "base", "exponent", "name", "symbol")

    base: int
    exponent: Numeric
    name: Optional[str]
    symbol: Optional[str]

    def __new__(
        cls,
        base: int,
        exponent: Numeric,
        name: Optional[str] = None,
        symbol: Optional[str] = None,
    ) -> "Prefix":
        if base != 0 and exponent == 0:
            return IdentityPrefix

        key = (base, exponent)
        if key in cls._known:
            return cls._known[key]

        self = super().__new__(cls)
        self._initialized = False
        cls._known[key] = self
        return self

    def __init__(
        self,
        base: int,
        exponent: Numeric,
        name: Optional[str] = None,
        symbol: Optional[str] = None,
    ) -> None:
        if self._initialized:
            return

        self.base = base
        self.exponent = exponent
        self.name = name
        self.symbol = symbol
        self._initialized = True

        if name:
            self._by_name[name] = self
        if symbol:
            self._by_symbol[symbol] = self

    @classmethod
    def resolve_symbol(cls, symbol: str) -> "Prefix":
        """Returns the Prefix with the given symbol"""
        return cls._by_symbol[symbol]

    # Pickle support

    def __getnewargs_ex__(self) -> Tuple[Tuple[int, Numeric], Dict[str, Any]]:
        return (self.base, self.exponent), {}

    # JSON support

    def __json__(self) -> Dict[str, Any]:
        return {
            "base": self.base,
            "exponent": self.exponent,
            "name": self.name,
            "symbol": self.symbol,
        }

    @classmethod
    def __from_json__(cls, json_object: Dict[str, Any]) -> "Prefix":
        return Prefix(json_object["base"], json_object["exponent"])

    # Pydantic support

    @classmethod
    def __get_validators__(
        cls,
    ) -> Generator[Callable[[Union[str, "Prefix"]], "Prefix"], None, None]:
        yield cls.validate

    @classmethod
    def validate(cls, value: Union[str, "Prefix"]) -> "Prefix":
        if isinstance(value, str):
            if value not in cls._by_name:
                raise ValueError(f"{value!r} is not a named Prefix")

            return cls._by_name[value]

        if isinstance(value, Prefix):
            return value

        raise ValueError(f"No conversion from {value!r} to Prefix")

    __repr__ = formatting.prefix_repr
    __str__ = formatting.prefix_str
    _repr_pretty_ = formatting.prefix_pretty
    _repr_html_ = formatting.mathml(formatting.prefix_mathml)

    def quantify(self) -> Numeric:
        return cast(Numeric, self.base**self.exponent)

    @overload
    def __mul__(self, other: "Prefix") -> "Prefix":
        ...  # pragma: no cover

    @overload
    def __mul__(self, other: "Unit") -> "Unit":
        ...  # pragma: no cover

    @overload
    def __mul__(self, other: Numeric) -> "Quantity":
        ...  # pragma: no cover

    def __mul__(
        self, other: Union["Prefix", "Unit", Numeric]
    ) -> Union["Prefix", "Unit", "Quantity"]:
        if isinstance(other, Prefix):
            if other.base == 0:
                return self
            elif self.base == 0:
                return other
            elif other.base == self.base:
                return Prefix(self.base, self.exponent + other.exponent)

            base, exponent = self.base, self.exponent
            exponent += other.exponent * (math.log(other.base) / math.log(self.base))

            return Prefix(base, exponent)

        if isinstance(other, Unit):
            return Unit(other.prefix * self, other.factors, other.dimension)

        if isinstance(other, NUMERIC_CLASSES):
            return (other * One) * self.quantify()

        return NotImplemented

    __rmul__ = __mul__

    def __truediv__(self, other: "Prefix") -> "Prefix":
        if not isinstance(other, Prefix):
            return NotImplemented

        if other.base == 0:
            return self
        elif self.base == 0:
            return Prefix(other.base, -other.exponent)
        elif other.base == self.base:
            return Prefix(self.base, self.exponent - other.exponent)

        base, exponent = self.base, self.exponent
        exponent -= other.exponent * (math.log(other.base) / math.log(self.base))

        return Prefix(base, exponent)

    def __pow__(self, power: int) -> "Prefix":
        return Prefix(self.base, self.exponent * power)

    def root(self, degree: int) -> "Prefix":
        """Returns the nth root of this Prefix"""
        if not isinstance(degree, int):
            raise TypeError(f"degree should be an integer, not {type(degree)}")

        if degree == 0:
            return IdentityPrefix

        if self.exponent // degree != self.exponent / degree:
            raise FractionalDimensionError(degree, self)

        return Prefix(self.base, int(self.exponent // degree))


class Unit:
    """Unit is a predetermined reference amount or definition for a measurable quantity

    `measured` includes a number of well-known units, and additional contributions are
    always welcome.  The SI system of units is covered in the [`measured.si`](../si)
    package.

    Attributes:

        name (Optional[str]): the name of this unit, which may not be set in the case
            of complex derived units

        symbol (Optional[str]): the symbol of this unit, which may not be set in the
            case of complex derived units

        prefix (Prefix): the prefix to multiply this unit by (`IdentityPrefix` for
            unprefixed units)

        factors (Mapping["Unit", int]): a mapping of the base units and their powers
            that make up a compound unit, or `{self: 1}` for base units

    Examples:

        >>> from measured import Length, Time, Speed
        >>> from measured.si import Meter, Second
        >>> Meter.name, Meter.symbol
        ('meter', 'm')

        >>> meter_per_second = Meter / Second
        >>> assert meter_per_second.dimension is Speed
        >>> assert meter_per_second.dimension is Length / Time
        >>> meter_per_second.name, meter_per_second.symbol
        (None, None)

        >>> back_to_meter = (meter_per_second * Second)
        >>> back_to_meter.name, back_to_meter.symbol
        ('meter', 'm')

        >>> assert Meter.factors == {Meter: 1}
        >>> assert Second.factors == {Second: 1}
        >>> assert meter_per_second.factors == {Meter: 1, Second: -1}


        [`Unit`][measured.Unit] values define format specifiers to control how they are
        formatted as strings:

        >>> from measured.si import Meter, Second
        >>> str(Meter / Second)
        'm⋅s⁻¹'
        >>> f"{Meter / Second}"
        'm⋅s⁻¹'
        >>> f"{Meter**2 / Second**2}"
        'm²⋅s⁻²'

        The `/` format specifier presents the unit as a ratio:

        >>> f"{Meter / Second:/}"
        'm/s'
        >>> f"{Meter**2/Second:/}"
        'm²/s'
        >>> f"{Meter**2/Second**2:/}"
        'm²/s²'
    """

    UnitKey = Tuple[Prefix, Tuple[Tuple["Unit", int], ...]]

    _known: ClassVar[Dict[UnitKey, "Unit"]] = {}
    _initialized: bool

    _base: ClassVar[Set["Unit"]] = set()
    _by_name: ClassVar[Dict[str, "Unit"]] = {}
    _by_symbol: ClassVar[Dict[str, "Unit"]] = {}

    __slots__ = ("_initialized", "prefix", "factors", "dimension", "names", "symbols")

    prefix: Prefix
    factors: Mapping["Unit", int]
    dimension: Dimension
    names: Tuple[str, ...]
    symbols: Tuple[str, ...]

    def __new__(
        cls,
        prefix: Prefix,
        factors: Mapping["Unit", int],
        dimension: Dimension,
        name: Optional[str] = None,
        symbol: Optional[str] = None,
    ) -> "Unit":
        key = cls._build_key(prefix, factors)
        if key in cls._known:
            return cls._known[key]

        if name and name in cls._by_name:
            return cls._by_name[name]

        self = super().__new__(cls)
        self._initialized = False
        if not factors:
            key = cls._build_key(prefix, {self: 1})
        cls._known[key] = self
        return self

    def __init__(
        self,
        prefix: Prefix,
        factors: Mapping["Unit", int],
        dimension: Dimension,
        name: Optional[str] = None,
        symbol: Optional[str] = None,
    ) -> None:
        if self._initialized:
            return

        self.prefix = prefix
        self.factors = factors or {self: 1}
        self.dimension = dimension
        self.names = tuple()
        self.symbols = tuple()
        self.alias(name=name, symbol=symbol)
        self._initialized = True

    @classmethod
    def _build_key(cls, prefix: Prefix, factors: Mapping["Unit", int]) -> UnitKey:
        factor_key = tuple(sorted(factors.items(), key=lambda pair: id(pair[0])))
        return (prefix, factor_key)

    @classmethod
    def base(cls) -> Iterable["Unit"]:
        return list(cls._base)

    @classmethod
    def define(cls, dimension: Dimension, name: str, symbol: str) -> "Unit":
        """Defines a new base unit"""
        if name in cls._by_name:
            raise ValueError(f"A unit named {name} is already defined")

        if symbol in cls._by_symbol:
            raise ValueError(f"A unit with symbol {symbol} is already defined")

        unit = cls(IdentityPrefix, {}, dimension, name, symbol)
        cls._base.add(unit)
        return unit

    @classmethod
    def derive(cls, unit: "Unit", name: str, symbol: str) -> "Unit":
        """Registers a new named unit derived from other units"""
        unit.alias(name=name, symbol=symbol)
        return unit

    @classmethod
    def named(cls, name: str) -> "Unit":
        """Returns the unit with the given name"""
        return cls._by_name[name]

    @classmethod
    def resolve_symbol(cls, symbol: str) -> "Unit":
        """Returns the unit with the given (possibly prefixed) symbol"""
        if symbol in cls._by_symbol:
            return cls._by_symbol[symbol]

        for i in range(1, len(symbol)):
            try:
                prefix = Prefix.resolve_symbol(symbol[:i])
                unit = cls._by_symbol[symbol[i:]]
            except KeyError:
                continue

            return prefix * unit

        if symbol in cls._by_name:
            return cls._by_name[symbol]

        raise KeyError(f"No unit (or prefixed unit) matching {symbol!r}")

    def equals(self, other: "Quantity") -> None:
        """Defines a conversion between this Unit and another"""
        if other.unit == self and self is not One:
            raise ValueError("No need to define conversions for a unit and itself")
        conversions.equate(1 * self, other)

    def alias(self, name: Optional[str] = None, symbol: Optional[str] = None) -> None:
        """Adds an alternative name and/or symbol to the unit"""
        if name:
            if name in self._by_name and self._by_name[name] is not self:
                raise ValueError(f"A unit named {name} is already defined")

            self.names = self.names + (name,)
            self._by_name[name] = self

        if symbol:
            if symbol in self._by_symbol and self._by_symbol[symbol] is not self:
                raise ValueError(f"A unit with symbol {symbol} is already defined")

            if symbol and " " in symbol:
                raise ValueError(f"{symbol!r} will not be parsable if it has spaces.")

            self.symbols = self.symbols + (symbol,)
            self._by_symbol[symbol] = self

    @property
    def name(self) -> Optional[str]:
        return self.names[0] if self.names else None

    @property
    def symbol(self) -> Optional[str]:
        return self.symbols[0] if self.symbols else None

    # Pickle support

    def __getnewargs_ex__(
        self,
    ) -> Tuple[Tuple[Prefix, Mapping["Unit", int], Dimension], Dict[str, Any]]:
        factors = {} if self.factors == {self: 1} else self.factors
        args = (self.prefix, factors, self.dimension)
        kwargs = {"name": self.name, "symbol": self.symbol}
        return args, kwargs

    # JSON support

    def __json__(self) -> Dict[str, Any]:
        prefix, factors = self._build_key(self.prefix, self.factors)
        return {
            "name": self.name,
            "symbol": self.symbol,
            "dimension": self.dimension,
            "prefix": prefix if prefix.quantify() != 1 else None,
            "factors": None if factors == ((self, 1),) else factors,
        }

    @classmethod
    def __from_json__(cls, json_object: Dict[str, Any]) -> "Unit":
        if not json_object["factors"]:
            return cls._by_name[json_object["name"]]
        prefix = json_object["prefix"] or Prefix(0, 0)
        factors = dict(json_object["factors"])
        dimension = json_object["dimension"]
        return Unit(prefix, factors, dimension)

    # Pydantic support

    @classmethod
    def __get_validators__(
        cls,
    ) -> Generator[Callable[[Union[str, "Unit"]], "Unit"], None, None]:
        yield cls.validate

    @classmethod
    def validate(cls, value: Union[str, "Unit"]) -> "Unit":
        if isinstance(value, str):
            if value not in cls._by_name:
                raise ValueError(f"{value!r} is not a named Unit")

            return cls._by_name[value]

        if isinstance(value, Unit):
            return value

        raise ValueError(f"No conversion from {value!r} to Unit")

    __repr__ = formatting.unit_repr
    __str__ = formatting.unit_str
    __format__ = formatting.unit_format
    _repr_pretty_ = formatting.unit_pretty
    _repr_html_ = formatting.mathml(formatting.unit_mathml)

    @classmethod
    def parse(cls, string: str) -> "Unit":
        """
        Parses a unit from a string.

        Examples:

            It is important to import any modules of units you will be parsing.  Units
            and their symbols are registered when they are first imported and created.

            >>> from measured import Unit, si

            Integer and floating point quantities can be parsed, along with units of
            any complexity.

            >>> meter_per_second = Unit.parse('m/s')
            >>> meter_per_second.dimension is Speed
            True
            >>> amperes_cubed_per_area = Unit.parse('A³/m²')
            >>> from measured.si import Meter, Ampere
            >>> amperes_cubed_per_area is Ampere**3/Meter**2
            True
            >>> str(amperes_cubed_per_area)
            'C³⋅s⁻³⋅m⁻²'

            Measured can parse any unit in the same format it produces with `str`, but
            also understands easier-to-type versions:

            >>> assert Unit.parse('m^2/s') == Unit.parse('m²⋅s⁻¹')
            >>> assert Unit.parse('m^2*s') == Unit.parse('m²⋅s')
        """
        return cast(Unit, parser.parse(string, start="unit"))

    @classmethod
    def _simplify(cls, factors: Mapping["Unit", int]) -> Dict["Unit", int]:
        simplified = {
            unit: exponent
            for unit, exponent in factors.items()
            if unit is not One and exponent != 0
        }
        return simplified or {One: 1}

    @lru_cache(maxsize=None)
    def quantify(self) -> "Quantity":
        """Produce a Quantity of this Unit, including the Prefix.

        Examples:

            >>> from measured.si import Kilo, Meter
            >>> assert Kilo * Meter != 1000 * Meter
            >>> assert 1 * Kilo * Meter == 1000 * Meter
            >>> assert (Kilo * Meter).quantify() == 1000 * Meter
        """
        return self.prefix.quantify() * Unit(
            IdentityPrefix, self.factors, self.dimension
        )

    def __add__(self, other: "Unit") -> "Unit":
        if self is not other:
            return NotImplemented

        return self

    def __sub__(self, other: "Unit") -> "Unit":
        if self is not other:
            return NotImplemented

        return self

    @staticmethod
    @lru_cache(maxsize=None)
    def _multiply(self: "Unit", other: "Unit") -> "Unit":
        dimension = self.dimension * other.dimension
        prefix = self.prefix * other.prefix

        factors: Dict["Unit", int] = defaultdict(int, self.factors)
        for unit, exponent in other.factors.items():
            factors[unit] += exponent
        factors = self._simplify(factors)

        return Unit(prefix, factors, dimension)

    @overload
    def __mul__(self, other: "Unit") -> "Unit":
        ...  # pragma: no cover

    @overload
    def __mul__(self, other: Numeric) -> "Quantity":
        ...  # pragma: no cover

    def __mul__(self, other: Union["Unit", Numeric]) -> Union["Unit", "Quantity"]:
        if isinstance(other, NUMERIC_CLASSES):
            return Quantity(other, self)

        if not isinstance(other, Unit):
            return NotImplemented

        return Unit._multiply(self, other)

    __rmul__ = __mul__

    @staticmethod
    @lru_cache(maxsize=None)
    def _divide(self: "Unit", other: "Unit") -> "Unit":
        dimension = self.dimension / other.dimension
        prefix = self.prefix / other.prefix

        factors: Dict["Unit", int] = defaultdict(int, self.factors)
        for unit, exponent in other.factors.items():
            factors[unit] -= exponent
        factors = self._simplify(factors)

        return Unit(prefix, factors, dimension)

    def __truediv__(self, other: "Unit") -> "Unit":
        if not isinstance(other, Unit):
            return NotImplemented

        return Unit._divide(self, other)

    def __pow__(self, power: int) -> "Unit":
        if not isinstance(power, int):
            return NotImplemented

        dimension = self.dimension**power
        prefix = self.prefix**power

        factors = self._simplify(
            {unit: exponent * power for unit, exponent in self.factors.items()}
        )
        return Unit(prefix, factors, dimension)

    def root(self, degree: int) -> "Unit":
        """Returns the nth root of this Unit"""
        if not isinstance(degree, int):
            raise TypeError(f"degree should be an integer, not {type(degree)}")

        if degree == 0:
            return One

        dimension = self.dimension.root(degree)
        prefix = self.prefix.root(degree)

        if any(
            exponent // degree != exponent / degree
            for unit, exponent in self.factors.items()
            if exponent > 0 and unit is not One
        ):
            raise FractionalDimensionError(degree, self)

        factors = self._simplify(
            {unit: int(exponent // degree) for unit, exponent in self.factors.items()}
        )
        return Unit(prefix, factors, dimension)

    @lru_cache(maxsize=None)
    def as_ratio(self) -> Tuple["Unit", "Unit"]:
        """Returns this unit, split into a numerator and denominator"""
        numerator, denominator = self.dimension.as_ratio()
        return (
            Unit(
                self.prefix,
                {u: e for u, e in self.factors.items() if e >= 0} or {One: 1},
                numerator,
            ),
            Unit(
                IdentityPrefix,
                {u: -e for u, e in self.factors.items() if e < 0} or {One: 1},
                denominator,
            ),
        )


@total_ordering
class Quantity:
    """Quantity represents a quantity of some Unit

    Attributes:

        magnitude (int | float): the quantity

        unit (Unit): the [`Unit`][measured.Unit]

    Examples:

        All of the arithmetic operations are supported between Quantities.
        Multiplication and division with [`Unit`][measured.Unit], `int`, and `float`
        are also supported.

        >>> from measured.si import Meter, Second
        >>> assert 2 * Meter + 3 * Meter == 5 * Meter
        >>> assert 3 * Meter - 1 * Meter == 2 * Meter
        >>> assert (2 * Meter) / (1 * Second) == 2 * Meter / Second
        >>> assert 10 * Meter / 2 == 5 * Meter
        >>> assert 10 * Meter * 2 == 20 * Meter
        >>> assert (10 * Meter)**2 == 100 * Meter**2

        Keep an eye on Python's operator precedence, which may lead to surprising
        results.  Consider the following:

        >>> assert (2 * Meter / 1 * Second) != (2 * Meter) / (1 * Second)

        In the above example, the order of operations on the left is:

        * `2 * Meter` → `Quantity(2, Meter)`
        * `Quantity(2, Meter) / 1` → `Quantity(2, Meter)`
        * `Quantity(2, Meter) * Second` → `Quantity(2, Meter * Second)`

        But on the right, the order is:

        * `2 * Meter` → `Quantity(2, Meter)`
        * `1 * Second` → `Quantity(1, Second)`
        * `Quantity(2, Meter) / Quantity(1, Second)` → `Quantity(2, Meter / Second)`

        [`Quantity`][measured.Quantity] values define format specifiers to control how
        they are formatted as strings:

        >>> from measured.si import Meter, Second
        >>> str(5 * Meter / Second)
        '5 m⋅s⁻¹'
        >>> f"{5 * Meter / Second}"
        '5 m⋅s⁻¹'
        >>> f"{5 * Meter**2 / Second**2}"
        '5 m²⋅s⁻²'

        When formatting a [`Quantity`][measured.Quantity], you can specify two separate
        format specifiers, separated by a `:`.  The first specifier is used to format
        the `magnitude`, and the second is used to format the `unit` (see
        [`Unit`][measured.Unit] for more information about the available specifiers).

        Specifying both the magnitude and unit format:

        >>> f"{5.1234 * Meter / Second:.2f:/}"
        '5.12 m/s'

        Specifying only the magnitude's format:

        >>> f"{5.1234 * Meter / Second:.2f}"
        '5.12 m⋅s⁻¹'

        Specifying only the unit's format:

        >>> f"{5.1234 * Meter / Second::/}"
        '5.1234 m/s'

    """

    __slots__ = ("magnitude", "unit")

    magnitude: Numeric
    unit: Unit

    def __init__(self, magnitude: Numeric, unit: Union[Unit, str]):
        self.magnitude = magnitude
        if isinstance(unit, str):
            unit = Unit.parse(unit)
        self.unit = unit

    def unprefixed(self) -> "Quantity":
        """Reduces this Quantity into a new Quantity expressed only in base units
        without any Prefixes"""
        return self.magnitude * self.unit.quantify()

    def in_unit(self, other: Unit) -> "Quantity":
        """Convert this Quantity into another unit"""
        return conversions.convert(self, other)

    # JSON support

    def __json__(self) -> Dict[str, Any]:
        return {
            "magnitude": self.magnitude,
            "unit": self.unit,
        }

    @classmethod
    def __from_json__(cls, json_object: Dict[str, Any]) -> "Quantity":
        return Quantity(json_object["magnitude"], json_object["unit"])

    # Pydantic support

    @classmethod
    def __get_validators__(
        cls,
    ) -> Generator[Callable[["Quantity"], "Quantity"], None, None]:
        yield cls.validate

    @classmethod
    def validate(cls, value: Union[str, "Quantity"]) -> "Quantity":
        if isinstance(value, Quantity):
            return value

        raise ValueError(f"No conversion from {value!r} to Quantity")

    # SQLAlchemy support

    def __composite_values__(self) -> Tuple[Numeric, str]:
        """
        From the [SQLAlchemy documentation][1]:

            The requirements for the custom datatype class are that it have a
            constructor which accepts positional arguments corresponding to its column
            format, and also provides a method __composite_values__() which returns the
            state of the object as a list or tuple, in order of its column-based
            attributes. It also should supply adequate __eq__() and __ne__() methods
            which test the equality of two instances.

            [1]: https://docs.sqlalchemy.org/en/20/orm/composites.html
        """
        return self.magnitude, str(self.unit)

    __repr__ = formatting.quantity_repr
    __str__ = formatting.quantity_str
    __format__ = formatting.quantity_format
    _repr_pretty_ = formatting.quantity_pretty
    _repr_html_ = formatting.mathml(formatting.quantity_mathml)

    @classmethod
    def parse(cls, string: str) -> "Quantity":
        """
        Parses a quantity from a string.

        Examples:

            It is important to import any modules of units you will be parsing.  Units
            and their symbols are registered when they are first imported and created.

            >>> from measured import Quantity, si

            Integer and floating point quantities can be parsed, along with units of
            any complexity.

            >>> distance = Quantity.parse('5200 m')
            >>> speed = Quantity.parse('5.2e2 m/s')
            >>> time = distance / speed
            >>> str(time)
            '10.0 s'
            >>> amperes_cubed_per_area = Quantity.parse('2 A³/m²')
            >>> str(amperes_cubed_per_area)
            '2 C³⋅s⁻³⋅m⁻²'

            Measured can parse any unit in the same format it produces with `str`, but
            also understands easier-to-type versions:

            >>> assert Quantity.parse('2 m^2/s') == Quantity.parse('2 m²⋅s⁻¹')
            >>> assert Quantity.parse('2 m^2*s') == Quantity.parse('2 m²⋅s')
        """
        return cast(Quantity, parser.parse(string, start="quantity"))

    def __hash__(self) -> int:
        return hash((self.magnitude, self.unit))

    def __add__(self, other: "Quantity") -> "Quantity":
        if isinstance(other, Quantity):
            other = other.in_unit(self.unit)
            return Quantity(self.magnitude + other.magnitude, self.unit)

        return NotImplemented

    def __sub__(self, other: "Quantity") -> "Quantity":
        if isinstance(other, Quantity):
            other = other.in_unit(self.unit)
            return Quantity(self.magnitude - other.magnitude, self.unit)

        return NotImplemented

    def __mul__(self, other: Union["Quantity", "Unit", Numeric]) -> "Quantity":
        if isinstance(other, Unit):
            return Quantity(self.magnitude, self.unit * other)

        if isinstance(other, Quantity):
            return Quantity(self.magnitude * other.magnitude, self.unit * other.unit)

        if isinstance(other, NUMERIC_CLASSES):
            return Quantity(self.magnitude * other, self.unit)

        return NotImplemented

    __rmul__ = __mul__

    def __truediv__(self, other: Union["Quantity", "Unit", Numeric]) -> "Quantity":
        if isinstance(other, Unit):
            return Quantity(self.magnitude, self.unit / other)

        if isinstance(other, Quantity):
            return Quantity(self.magnitude / other.magnitude, self.unit / other.unit)

        if isinstance(other, NUMERIC_CLASSES):
            return Quantity(self.magnitude / other, self.unit)

        return NotImplemented

    def __rtruediv__(self, other: Numeric) -> "Quantity":
        if isinstance(other, NUMERIC_CLASSES):
            return Quantity(other / self.magnitude, self.unit)

        return NotImplemented

    def __pow__(self, power: int) -> "Quantity":
        return Quantity(self.magnitude**power, self.unit**power)

    def root(self, degree: int) -> "Quantity":
        """Returns the nth root of this Quantity"""
        if degree == 0:
            return 1 * One
        return Quantity(self.magnitude ** (1 / degree), self.unit.root(degree))

    def __neg__(self) -> "Quantity":
        return Quantity(-self.magnitude, self.unit)

    def __pos__(self) -> "Quantity":
        return Quantity(+self.magnitude, self.unit)

    def __abs__(self) -> "Quantity":
        return Quantity(abs(self.magnitude), self.unit)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Level):
            other = other.quantify()

        if not isinstance(other, Quantity):
            return NotImplemented

        if self.unit.dimension != other.unit.dimension:
            return NotImplemented

        this = self.unprefixed()
        other = other.unprefixed()

        if this.unit == other.unit:
            return this.magnitude == other.magnitude

        try:
            return this.in_unit(other.unit) == other
        except conversions.ConversionNotFound:
            return NotImplemented

    def __lt__(self, other: Any) -> bool:
        if isinstance(other, Level):
            other = other.quantify()

        if not isinstance(other, Quantity):
            return NotImplemented

        if self.unit.dimension != other.unit.dimension:
            return NotImplemented

        this = self.unprefixed()
        other = other.unprefixed()

        if this.unit == other.unit:
            return this.magnitude < other.magnitude

        try:
            return this.in_unit(other.unit) < other
        except conversions.ConversionNotFound:
            return NotImplemented

    def level(self, unit: "LogarithmicUnit") -> "Level":
        return unit.level(self)


class Logarithm:
    """A `Logarithm` forms a family of [`LogarithmicUnits`][measured.LogarithmicUnit],
    which measures the _ratio_ of a measured [`Quantity`][measured.Quantity] to a
    reference [`Quantity`][measured.Quantity] on a logarithmic scale.  Commonly used
    logarithms include the `Decibel` (base 10) and the `Neper` (base _e_)

    The logarithm is not a unit itself, but rather a family of units, each with a
    specific reference [`Quantity`][measured.Quantity].  For example, the unit `dBW`,
    commonly used in electrical and communications applications, measures the ratio of a
    `Power` quantity against the reference value 1 Watt, in increments of 10 (the base)
    to the 1/10th power (the deci- prefix).

    Examples:

        Use a `Logarithm`, like `Decibel`, to _create_ a new
        [`LogarithmicUnit`][measured.LogarithmicUnit] based on a reference value.  In
        the example below, the reference value is 1 Watt:

        >>> from measured import Decibel
        >>> from measured.si import Watt
        >>> dBW = Decibel[1 * Watt]

        The `dBW` unit may now be used to produces [`Levels`][measured.Level] that may
        be used in operations with Quantities of the same dimension (`Power`, in this
        case):

        >>> assert 100 * Watt == 20 * dBW
    """

    _known: ClassVar[Dict[Tuple[float, Prefix], "Logarithm"]] = {}
    _initialized: bool

    base: float
    prefix: Prefix
    name: Optional[str]
    symbol: Optional[str]

    def __new__(
        cls,
        base: float,
        prefix: Prefix = Prefix(0, 0),
        name: Optional[str] = None,
        symbol: Optional[str] = None,
    ) -> "Logarithm":
        key = (base, prefix)
        if key in cls._known:
            return cls._known[key]

        self = super().__new__(cls)
        self._initialized = False
        cls._known[key] = self
        return self

    def __init__(
        self,
        base: float,
        prefix: Prefix = Prefix(0, 0),
        name: Optional[str] = None,
        symbol: Optional[str] = None,
    ) -> None:
        if self._initialized:
            return

        self.base = base
        self.prefix = prefix
        self.name = name
        self.symbol = symbol
        self._initialized = True

    def alias(self, name: str, symbol: str) -> "Logarithm":
        """Gives a name and symbol to this Logarithm"""
        self.name = name
        self.symbol = symbol
        return self

    __repr__ = formatting.logarithm_repr
    __str__ = formatting.logarithm_str
    _repr_pretty_ = formatting.logarithm_pretty
    _repr_html_ = formatting.mathml(formatting.logarithm_mathml)

    def __mul__(self, other: Prefix) -> "Logarithm":
        if isinstance(other, Prefix):
            return Logarithm(self.base, prefix=self.prefix * other)

        return NotImplemented

    __rmul__ = __mul__

    def __getitem__(self, reference: Quantity) -> "LogarithmicUnit":
        return LogarithmicUnit(self, reference)


class LogarithmicUnit:
    """A `LogarithmicUnit` represents a _ratio_ of a measured
    [`Quantity`][measured.Quantity] to a reference [`Quantity`][measured.Quantity] on a
    logarithmic scale.

    Examples:

        You may construct a `LogarithmicUnit` by applying the reference quantity as a
        "suffix" (using Python's indexing operator `[]`) to the
        [`Logarithm`][measured.Logarithm] in question:

        >>> from measured import Decibel
        >>> from measured.si import Watt, Milli
        >>> dBW = Decibel[1 * Watt]
        >>> dBm = Decibel[1 * Milli * Watt]
        >>> assert dBm is not dBW

        `LogarithmicUnits` create [`Levels`][measured.Level] through multiplication:

        >>> level = 20 * dBW

        A [`Level`][measured.Level] is analogous to a [`Quantity`][measured.Quantity]:

        >>> level.magnitude
        20
        >>> level.unit is dBW
        True
        >>> 100 * Watt == level
        True
    """

    _known: ClassVar[Dict[Tuple[Logarithm, Quantity], "LogarithmicUnit"]] = {}
    _initialized: bool

    logarithm: Logarithm
    reference: Quantity
    name: Optional[str]
    symbol: Optional[str]

    def __new__(
        cls,
        logarithm: Logarithm,
        reference: Quantity,
        name: Optional[str] = None,
        symbol: Optional[str] = None,
    ) -> "LogarithmicUnit":
        key = (logarithm, reference)
        if key in cls._known:
            return cls._known[key]

        self = super().__new__(cls)
        self._initialized = False
        cls._known[key] = self
        return self

    def __init__(
        self,
        logarithm: Logarithm,
        reference: Quantity,
        name: Optional[str] = None,
        symbol: Optional[str] = None,
    ) -> None:
        if self._initialized:
            return

        self.logarithm = logarithm
        self.reference = reference.unprefixed()
        self.name = name
        self.symbol = symbol
        self._initialized = True

    def alias(self, name: str, symbol: str) -> "LogarithmicUnit":
        """Gives a name and symbol to this LogarithmicUnit"""
        assert not self.name and not self.symbol
        self.name = name
        self.symbol = symbol
        return self

    @property
    def power_ratio(self) -> int:
        """Indicates whether the reference quantity is a power (1) or root-power (2)
        quantity"""
        if self.reference.unit.dimension in ROOT_POWER_DIMENSIONS:
            return 2
        return 1

    __repr__ = formatting.logarithmic_unit_repr
    __str__ = formatting.logarithmic_unit_str
    _repr_pretty_ = formatting.logarithmic_unit_pretty
    _repr_html_ = formatting.mathml(formatting.logarithmic_unit_mathml)

    def __mul__(self, magnitude: Numeric) -> "Level":
        return Level(magnitude, self)

    __rmul__ = __mul__

    def level(self, quantity: Quantity) -> "Level":
        base = self.logarithm.base
        prefix = self.logarithm.prefix
        power_ratio = self.power_ratio
        ratio = quantity.in_unit(self.reference.unit) / self.reference
        magnitude = (
            power_ratio * (1 / prefix.quantify()) * math.log(ratio.magnitude, base)
        )
        return Level(magnitude, self)


class Level:
    """A `Level` is analogous to a [`Quantity`][measured.Quantity], but represents a
    ratio of a quantity to a reference quantity.  The reference quantity is captured by
    the [`LogarithmicUnit`][measured.LogarithmicUnit].

    Attributes:

        magnitude (int | float): the quantity

        unit (LogarithmicUnit): the [`LogarithmicUnit`][measured.LogarithmicUnit]
    """

    magnitude: Numeric
    unit: LogarithmicUnit

    def __init__(self, magnitude: Numeric, unit: LogarithmicUnit) -> None:
        self.magnitude = magnitude
        self.unit = unit

    __repr__ = formatting.level_repr
    __str__ = formatting.level_str
    _repr_pretty_ = formatting.level_pretty
    _repr_html_ = formatting.mathml(formatting.level_mathml)

    def quantify(self) -> Quantity:
        """Converts this Level into a Quantity of the reference unit"""
        base = self.unit.logarithm.base
        prefix = self.unit.logarithm.prefix
        power_ratio = self.unit.power_ratio
        reference = self.unit.reference
        exponent = self.magnitude * prefix.quantify()
        magnitude: float = base ** (exponent / power_ratio)
        return magnitude * reference

    def __add__(self, other: "Level") -> "Level":
        if isinstance(other, Level):
            if other.unit != self.unit:
                return NotImplemented

            base = self.unit.logarithm.base
            prefix = self.unit.logarithm.prefix
            power_ratio = self.unit.power_ratio

            left_exponent = self.magnitude * prefix.quantify()
            right_exponent = other.magnitude * prefix.quantify()

            left_magnitude: float = base ** (left_exponent * power_ratio)
            right_magnitude: float = base ** (right_exponent * power_ratio)

            magnitude = base * math.log(left_magnitude + right_magnitude, base)

            return Level(magnitude, self.unit)

        return NotImplemented

    def __sub__(self, other: "Level") -> "Level":
        if isinstance(other, Level):
            if other.unit != self.unit:
                return NotImplemented

            base = self.unit.logarithm.base
            prefix = self.unit.logarithm.prefix
            power_ratio = self.unit.power_ratio

            left_exponent = self.magnitude * prefix.quantify()
            right_exponent = other.magnitude * prefix.quantify()

            left_magnitude: float = base ** (left_exponent * power_ratio)
            right_magnitude: float = base ** (right_exponent * power_ratio)

            magnitude = base * math.log(left_magnitude - right_magnitude, base)

            return Level(magnitude, self.unit)

        return NotImplemented

    def __mul__(self, other: Numeric) -> "Level":
        if isinstance(other, NUMERIC_CLASSES):
            return Level(self.magnitude + other, self.unit)

        return NotImplemented

    def __truediv__(self, other: Numeric) -> "Level":
        if isinstance(other, NUMERIC_CLASSES):
            return Level(self.magnitude - other, self.unit)

        return NotImplemented

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Level):
            return self.quantify() == other.quantify()

        if isinstance(other, Quantity):
            return self.quantify() == other

        return NotImplemented


class Measurement:
    """Measurement represents an uncertain measurement of some Quantity, and will
    propagate that uncertainty through arithmetic operations with Quantities and other
    Measurements.

    Attributes:

        measurand (Quantity): the measured value

        uncertainty (Quantity | Numeric): the uncertainty of the measurement

    Examples:

        All of the arithmetic operations are supported with Quantities and Measurements:

        >>> from measured import Measurement
        >>> from measured.si import Meter, Second
        >>> length = Measurement(2 * Meter, uncertainty=0.1)
        >>> width = Measurement(3 * Meter, uncertainty=0.2)
        >>> depth = 4 * Meter

        The uncertainty of the measurements propagate through arithmetic operations
        assuming that they are Gaussian, normally-distributed errors.

        >>> area = length * width
        >>> assert isinstance(area, Measurement)
        >>> assert area.measurand == 6 * Meter**2
        >>> assert area.uncertainty == 0.5 * Meter**2

        Quantities are treated as if they have an uncertainty of 0:

        >>> volume = area * depth
        >>> assert isinstance(volume, Measurement)
        >>> assert volume.measurand == 24 * Meter**3
        >>> assert volume.uncertainty == 2.0 * Meter**3

        When formatting a [`Measurement`][measured.Measurement], you can specify up to
        three separate format specifiers, separated by a `:`.  The first specifier
        controls how the uncertainty will be formatted.

        >>> speed = Measurement(10.0 * Meter / Second, 0.1)

        Uncertainties can be formatted as a ± interval using either the "±" or "+"
        specifier (the default)

        >>> f"{speed}"
        '10.0±0.1 m⋅s⁻¹'

        They can also be formatted as percentages using the "%" specifier:

        >>> f"{speed:%}"
        '10.0±1.00% m⋅s⁻¹'

        In any of those cases, the significant digits of the uncertainty can also be
        specified:

        >>> f"{speed:%.4f}"
        '10.0±1.0000% m⋅s⁻¹'
        >>> f"{speed:±.4f}"
        '10.0±0.1000 m⋅s⁻¹'
        >>> f"{speed:+.4f}"
        '10.0±0.1000 m⋅s⁻¹'

        The second and third specifiers are used to format the measurand and follow the
        same conventions as with [`Quantity`][measured.Quantity].

        >>> f"{speed:%.4f:.3f:/}"
        '10.000±1.0000% m/s'
    """

    measurand: Quantity
    uncertainty: Quantity

    def __init__(
        self, measurand: Quantity, uncertainty: Union[Numeric, Quantity]
    ) -> None:
        self.measurand = measurand
        if not isinstance(uncertainty, Quantity):
            uncertainty = Quantity(uncertainty, measurand.unit)
        assert uncertainty.unit is measurand.unit
        self.uncertainty = abs(uncertainty)

    @property
    def uncertainty_ratio(self) -> float:
        """The uncertainty, expressed as a fraction of the magnitude of the measurand"""
        return self.uncertainty.magnitude / self.measurand.magnitude

    @property
    def uncertainty_percent(self) -> float:
        """The uncertainty, expressed as a percent of the magnitude of the measurand"""
        return self.uncertainty_ratio * 100

    __repr__ = formatting.measurement_repr
    __str__ = formatting.measurement_str
    __format__ = formatting.measurement_format
    _repr_pretty_ = formatting.measurement_pretty
    _repr_html_ = formatting.mathml(formatting.measurement_mathml)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Quantity):
            other = Measurement(other, 0)
        if isinstance(other, Level):
            other = Measurement(other.quantify(), 0)

        if not isinstance(other, Measurement):
            return False

        if self.measurand.unit.dimension is not other.measurand.unit.dimension:
            return False

        self_lower = self.measurand - self.uncertainty
        other_lower = other.measurand - other.uncertainty
        self_upper = self.measurand + self.uncertainty
        other_upper = other.measurand + other.uncertainty

        try:
            overlaps_lower = self_lower <= other_lower <= self_upper
            overlaps_upper = self_lower <= other_upper <= self_upper
        except TypeError:
            return False

        return overlaps_lower or overlaps_upper

    def __lt__(self, other: object) -> bool:
        if isinstance(other, Quantity):
            other = Measurement(other, 0)
        if isinstance(other, Level):
            other = Measurement(other.quantify(), 0)

        if not isinstance(other, Measurement):
            return NotImplemented

        self_lower = self.measurand - self.uncertainty
        other_lower = other.measurand - other.uncertainty
        return self_lower < other_lower

    def __le__(self, other: object) -> bool:
        if isinstance(other, Quantity):
            other = Measurement(other, 0)
        if isinstance(other, Level):
            other = Measurement(other.quantify(), 0)

        if not isinstance(other, Measurement):
            return NotImplemented

        self_lower = self.measurand - self.uncertainty
        other_lower = other.measurand - other.uncertainty
        return self_lower <= other_lower

    def __gt__(self, other: object) -> bool:
        if isinstance(other, Quantity):
            other = Measurement(other, 0)
        if isinstance(other, Level):
            other = Measurement(other.quantify(), 0)

        if not isinstance(other, Measurement):
            return NotImplemented

        self_upper = self.measurand + self.uncertainty
        other_upper = other.measurand + other.uncertainty
        return self_upper > other_upper

    def __ge__(self, other: object) -> bool:
        if isinstance(other, Quantity):
            other = Measurement(other, 0)
        if isinstance(other, Level):
            other = Measurement(other.quantify(), 0)

        if not isinstance(other, Measurement):
            return NotImplemented

        self_upper = self.measurand + self.uncertainty
        other_upper = other.measurand + other.uncertainty
        return self_upper >= other_upper

    def __add__(self, other: Union["Measurement", Quantity]) -> "Measurement":
        if isinstance(other, Quantity):
            other = Measurement(other, 0)

        if not isinstance(other, Measurement):
            return NotImplemented

        measurand = self.measurand + other.measurand
        uncertainty = (self.uncertainty**2 + other.uncertainty**2).root(2)
        return Measurement(measurand, uncertainty)

    __radd__ = __add__

    def __sub__(self, other: Union["Measurement", Quantity]) -> "Measurement":
        if isinstance(other, Quantity):
            other = Measurement(other, 0)

        if not isinstance(other, Measurement):
            return NotImplemented

        measurand = self.measurand - other.measurand
        uncertainty = (self.uncertainty**2 + other.uncertainty**2).root(2)
        return Measurement(measurand, uncertainty)

    def __rsub__(self, other: Union["Measurement", Quantity]) -> "Measurement":
        if isinstance(other, Quantity):
            other = Measurement(other, 0)

        if not isinstance(other, Measurement):
            return NotImplemented

        return other - self

    def __mul__(self, other: Union["Measurement", Quantity]) -> "Measurement":
        if isinstance(other, Quantity):
            other = Measurement(other, 0)

        if not isinstance(other, Measurement):
            return NotImplemented

        measurand = self.measurand * other.measurand
        uncertainty = math.sqrt(
            measurand.magnitude**2
            * (
                (self.uncertainty.magnitude**2 / self.measurand.magnitude**2)
                + (other.uncertainty.magnitude**2 / other.measurand.magnitude**2)
            )
        )
        return Measurement(measurand, uncertainty)

    __rmul__ = __mul__

    def __truediv__(self, other: Union["Measurement", Quantity]) -> "Measurement":
        if isinstance(other, Quantity):
            other = Measurement(other, 0)

        if not isinstance(other, Measurement):
            return NotImplemented

        measurand = self.measurand / other.measurand
        uncertainty = math.sqrt(
            measurand.magnitude**2
            * (
                (self.uncertainty.magnitude**2 / self.measurand.magnitude**2)
                + (other.uncertainty.magnitude**2 / other.measurand.magnitude**2)
            )
        )
        return Measurement(measurand, uncertainty)

    def __rtruediv__(self, other: Union["Measurement", Quantity]) -> "Measurement":
        if isinstance(other, Quantity):
            other = Measurement(other, 0)

        if not isinstance(other, Measurement):
            return NotImplemented

        return other / self

    def __pow__(self, exponent: int) -> "Measurement":
        if not isinstance(exponent, int):
            return NotImplemented

        measurand = self.measurand**exponent
        uncertainty = math.sqrt(
            (exponent * self.measurand.magnitude**2 * self.uncertainty.magnitude) ** 2
        )
        return Measurement(measurand, uncertainty)


def approximately(
    quantity: Union[Quantity, Level], within: float = 1e-7
) -> Measurement:
    """
    An approximation used mostly for making test assertions.  This special type of
    measurement uses a relative value for uncertainty and overrides its repr to be
    more useful for test assertions.

    Parameters:

        quantity (Quantity): the quantity to approximate

        within (float): An uncertainty of the measurement relative to the quantity

    Examples:

        >>> from measured import approximately
        >>> from measured.si import Meter
        >>> assert 5.2 * Meter == approximately(5 * Meter, 0.3)
        >>> assert 5.2 * Meter < approximately(6 * Meter, 0.5)
    """
    if isinstance(quantity, Level):
        quantity = quantity.quantify()

    return Measurement(quantity, uncertainty=(quantity.magnitude or 1.0) * within)


# https://en.wikipedia.org/wiki/Dimensional_analysis#Definition

# Fundamental physical dimensions

Number = Dimension.define(name="number", symbol="1")
Length = Dimension.define(name="length", symbol="L")
Time = Dimension.define(name="time", symbol="T")
Mass = Dimension.define(name="mass", symbol="M")
Temperature = Dimension.define(name="temperature", symbol="Θ")
Charge = Dimension.define(name="charge", symbol="Q")
AmountOfSubstance = Dimension.define(name="amount of substance", symbol="N")
LuminousIntensity = Dimension.define(name="luminous intensity", symbol="J")
Information = Dimension.define(name="information", symbol="B")  # TODO


# Derived dimensions

Area = Dimension.derive(Length * Length, name="area")
Volume = Dimension.derive(Area * Length, name="volume")
VolumetricFlow = Dimension.derive(Volume / Time, name="flow")

PlaneAngle = Length / Length
SolidAngle = Area / Area

# https://en.wikipedia.org/wiki/Fourth,_fifth,_and_sixth_derivatives_of_position

Speed = Dimension.derive(Length / Time, name="speed")
Acceleration = Dimension.derive(Length / Time**2, name="acceleration")
Jerk = Dimension.derive(Length / Time**3, name="jerk")
Snap = Dimension.derive(Length / Time**4, name="snap")
Crackle = Dimension.derive(Length / Time**5, name="crackle")
Pop = Dimension.derive(Length / Time**6, name="pop")

Frequency = Dimension.derive(Number / Time, name="frequency")

# https://en.wikipedia.org/wiki/Newton%27s_laws_of_motion#Second

Force = Dimension.derive(Mass * Acceleration, name="force")
Pressure = Dimension.derive(Force / Area, name="pressure")
Energy = Dimension.derive(Length * Force, name="energy")
Power = Dimension.derive(Energy / Time, name="power")

Current = Dimension.derive(Charge / Time, name="current")
Potential = Dimension.derive(Energy / Charge, name="potential")
Capacitance = Dimension.derive(Charge / Potential, name="capacitance")
Resistance = Dimension.derive(Potential / Current, name="resistance")
Conductance = Dimension.derive(Current / Potential, name="conductance")
Inductance = Dimension.derive(Potential * Time / Current, name="inductance")

MagneticFlux = Dimension.derive(Potential * Time, name="magnetic flux")
MagneticBField = Dimension.derive(Potential * Time / Area, name="magnetic B-field")


LuminousFlux = LuminousIntensity * SolidAngle
Illuminance = Dimension.derive(LuminousIntensity / Area, name="illuminance")

RadioactiveDose = Dimension.derive(Power / Mass, name="radioactivedose")

Catalysis = Dimension.derive(AmountOfSubstance / Time, name="catalysis")


# Fundamental prefixes

IdentityPrefix = Prefix(0, 0)


# Fundamental units

One = Number.unit(name="one", symbol="1")


from . import conversions  # noqa: E402
from .parsing import parser  # noqa: E402

One.equals(1 * One)


# Logarithms
# https://en.wikipedia.org/wiki/Decibel
# https://en.wikipedia.org/wiki/Neper

Neper = Logarithm(base=math.e, name="neper", symbol="Np")
Octave = Logarithm(base=2, name="octave", symbol="oct")
Bel = Logarithm(base=10, name="bel", symbol="bel")
Decibel = (Prefix(10, -1) * Bel).alias(name="decibel", symbol="dB")

# https://en.wikipedia.org/wiki/Power,_root-power,_and_field_quantities
ROOT_POWER_DIMENSIONS = {
    Potential,
    Current,
    Pressure,
    Potential / Length,
    Speed,
    Charge / Length,
    Charge / Area,
    Charge / Volume,
}



"""
From the [SQLAlchemy documentation][1]:

    The requirements for the custom datatype class are that it have a constructor which
    accepts positional arguments corresponding to its column format, and also provides a
    method __composite_values__() which returns the state of the object as a list or
    tuple, in order of its column-based attributes. It also should supply adequate
    __eq__() and __ne__() methods which test the equality of two instances.

    [1]: https://docs.sqlalchemy.org/en/20/orm/composites.html
"""
import pytest

pytest.importorskip("sqlalchemy")

# flake8: noqa: E402 (imports not at top of file)

from typing import Generator, Tuple

from sqlalchemy import Column, Float, Integer, String, create_engine, select
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session, composite, declarative_base

from measured import Numeric, Quantity
from measured.si import Kelvin, Meter, Ohm, Second
from measured.us import Fahrenheit, Inch


@pytest.mark.parametrize(
    "quantity, values",
    [
        (5 * Meter, (5, "m")),
        (5.0 * Meter, (5.0, "m")),
        (5.0 * Ohm, (5.0, "Ω")),
        (5.0 * Meter / Second, (5.0, "m⋅s⁻¹")),
        (5.0 * Meter**2 / Second, (5.0, "m²⋅s⁻¹")),
    ],
)
def test_quantity_composite_values_roundtrip(
    quantity: Quantity, values: Tuple[Numeric, str]
) -> None:
    assert quantity.__composite_values__() == values
    assert Quantity(*quantity.__composite_values__()) == quantity


Base = declarative_base()


class Measurement(Base):  # type: ignore
    __tablename__ = "measurements"

    id = Column(Integer, primary_key=True)

    temperature_magnitude = Column(Integer)
    temperature_unit = Column(String)
    temperature = composite(Quantity, temperature_magnitude, temperature_unit)

    area_magnitude = Column(Float)
    area_unit = Column(String)
    area = composite(Quantity, area_magnitude, area_unit)


@pytest.fixture
def engine() -> Generator[Engine, None, None]:
    engine = create_engine("sqlite://", future=True)
    Base.metadata.create_all(engine)
    try:
        yield engine
    finally:
        engine.dispose()


@pytest.fixture
def session(engine: Engine) -> Generator[Session, None, None]:
    with Session(engine) as session:
        yield session


def test_saving_and_reading_quantities(session: Session) -> None:
    first = Measurement(id=1, temperature=110 * Kelvin, area=10.5 * Meter**2)
    second = Measurement(id=2, temperature=350 * Fahrenheit, area=22.47 * Inch**2)
    session.add_all([first, second])
    session.flush()

    saved = set(session.query(Measurement))

    assert saved == {first, second}
    assert {m.temperature for m in saved} == {110 * Kelvin, 350 * Fahrenheit}
    assert {m.area for m in saved} == {10.5 * Meter**2, 22.47 * Inch**2}


def test_querying_for_quantities_exactly(session: Session) -> None:
    first = Measurement(id=1, temperature=110 * Kelvin, area=10.5 * Meter**2)
    second = Measurement(id=2, temperature=350 * Fahrenheit, area=22.47 * Inch**2)
    session.add_all([first, second])
    session.flush()

    found = set(
        session.query(Measurement).filter(Measurement.temperature == 110 * Kelvin)
    )

    assert found == {first}



from hypothesis.strategies import (
    SearchStrategy,
    builds,
    floats,
    integers,
    one_of,
    sampled_from,
)

from measured import systems  # noqa: F401
from measured import Dimension, Prefix, Quantity, Unit

DIMENSIONS = sorted(Dimension.fundamental(), key=lambda d: d.exponents)
PREFIXES = sorted(Prefix._by_name.values(), key=lambda p: (p.base, p.exponent))
BASE_UNITS = sorted(Unit.base(), key=lambda u: u.name or "")
UNITS = sorted(Unit._known.values(), key=lambda u: u.name or "")
for this in BASE_UNITS:
    UNITS.append(this**2)
    UNITS.append(this**3)
    UNITS.append(this**-2)
    UNITS.append(this**-3)
    for other in BASE_UNITS:
        UNITS.append(this / other)
        UNITS.append(this * other)
UNITS_WITH_SYMBOLS = [u for u in UNITS if u.symbol]


def dimensions() -> SearchStrategy[Dimension]:
    return sampled_from(DIMENSIONS)


def prefixes() -> SearchStrategy[Prefix]:
    return sampled_from(PREFIXES)


def units() -> SearchStrategy[Unit]:
    return sampled_from(UNITS)


def base_units() -> SearchStrategy[Unit]:
    return sampled_from(BASE_UNITS)


def units_with_symbols() -> SearchStrategy[Unit]:
    return sampled_from(UNITS_WITH_SYMBOLS)


def quantities() -> SearchStrategy[Quantity]:
    return builds(
        Quantity,
        magnitude=one_of(
            floats(min_value=-1e30, max_value=-1e-30),
            sampled_from([0.0]),
            floats(min_value=1e-30, max_value=1e-30),
            integers(),
        ),
        unit=units(),
    )



from measured import approximately
from measured.eu import Gradian
from measured.geometry import π
from measured.si import Arcminute, Degree, Radian


def test_gradian() -> None:
    assert 1 * Gradian == (π / 200 * Radian)
    assert 1 * Gradian == (9 / 10 * Degree)
    assert 1 * Gradian == approximately(54 * Arcminute)



from measured import (  # noqa: F401
    acoustics,
    apocrypha,
    astronomical,
    avoirdupois,
    computing,
    electronics,
    energy,
    eu,
    fff,
    iec,
    iso,
    metric,
    music,
    natural,
    si,
    troy,
    us,
)



import pytest

from measured import Charge, Length, Mass, Quantity, Time, approximately
from measured.natural import NaturalCharge, NaturalLength, NaturalMass, NaturalTime
from measured.physics import c, mₑ, ε0, ℏ
from measured.si import Coulomb, Kilogram, Meter, Second


@pytest.mark.parametrize(
    "fundamental_constant, unity",
    [
        (c, 1 * (NaturalLength / NaturalTime)),
        (mₑ, 1 * NaturalMass),
        (ℏ, 1 * ((NaturalLength**2 * NaturalMass) / NaturalTime)),
        (
            ε0,
            (
                1
                * (
                    (NaturalTime**2 * NaturalCharge**2)
                    / (NaturalMass * NaturalLength**3)
                )
            ),
        ),
    ],
)
def test_unity(fundamental_constant: Quantity, unity: Quantity) -> None:
    assert fundamental_constant == approximately(unity)
    assert unity == approximately(fundamental_constant)


def test_length() -> None:
    assert NaturalLength.dimension is Length
    assert 1 * NaturalLength == approximately(3.862e-13 * Meter, within=2e-4)


def test_mass() -> None:
    assert NaturalMass.dimension is Mass
    assert 1 * NaturalMass == approximately(9.109e-31 * Kilogram, within=5e-5)


def test_time() -> None:
    assert NaturalTime.dimension is Time
    assert 1 * NaturalTime == approximately(1.288e-21 * Second, within=7e-5)


def test_charge() -> None:
    assert NaturalCharge.dimension is Charge
    assert 1 * NaturalCharge == approximately(5.291e-19 * Coulomb, within=4e-5)



from measured import Acceleration, Force, Pressure, approximately
from measured.physics import gₙ
from measured.si import Kilo, Newton, Pascal, Second
from measured.us import PSI, Foot, GForce, Pound, PoundForce


def test_g_force() -> None:
    assert GForce.name == "g-force"
    assert GForce.symbol == "g-force"
    assert GForce.dimension is Acceleration
    assert 1 * GForce == gₙ
    assert 1 * GForce == approximately(32.17405 * Foot / Second**2, 5e-8)


def test_poundforce() -> None:
    assert PoundForce.name == "pound-force"
    assert PoundForce.symbol == "lbf"
    assert PoundForce.dimension is Force
    assert 1 * PoundForce == 1 * Pound * gₙ
    assert 1 * PoundForce == 1 * Pound * GForce
    assert 1 * PoundForce == 4.4482216152605 * Newton


def test_psi() -> None:
    assert PSI.name == "pounds per square inch"
    assert PSI.symbol == "psi"
    assert PSI.dimension is Pressure
    assert 1 * PSI == approximately(6894.757 * Pascal, 5e-8)
    assert 1 * PSI == approximately(6.894757 * Kilo * Pascal, 5e-8)



from measured import approximately
from measured.astronomical import (
    H0,
    AstronomicalUnit,
    Crab,
    EarthMass,
    HubbleLength,
    HubbleTime,
    HubbleVolume,
    Jansky,
    JulianYear,
    JupiterMass,
    LightYear,
    Parsec,
    SiderealDay,
    Siriometer,
    SolarMass,
)
from measured.energy import Erg
from measured.geometry import π
from measured.metric import Barn
from measured.physics import G
from measured.si import (
    Atto,
    Centi,
    ElectronVolt,
    Hertz,
    Hour,
    Kilo,
    Liter,
    Mega,
    Meter,
    Milli,
    Second,
    Watt,
)
from measured.us import Inch


def test_astronomical_unit() -> None:
    assert 1 * AstronomicalUnit == approximately(1.495978707e11 * Meter)
    assert 1 * Siriometer == approximately(1e6 * AstronomicalUnit)


def test_parsec() -> None:
    # https://en.wikipedia.org/wiki/Parsec#Calculating_the_value_of_a_parsec
    assert π * Parsec == approximately(180 * 60 * 60 * AstronomicalUnit)
    assert 1 * Parsec == approximately(3.0856775814913673e16 * Meter)


def test_light_year() -> None:
    # https://en.wikipedia.org/wiki/Light-year
    assert 1 * LightYear == approximately(9460730472580800 * Meter)
    assert 1 * LightYear == approximately(63241.077 * AstronomicalUnit)
    assert 1 * LightYear == approximately(0.306601 * Parsec, within=2e-6)


def test_attoparsec() -> None:
    assert 1 * Atto * Parsec == approximately(3.086 * Centi * Meter, within=2e-4)
    assert 1 * Atto * Parsec == approximately(1.215 * Inch, within=2e-4)


def test_barn_megaparsec() -> None:
    assert 1 * Barn * (Mega * Parsec) == approximately(3 * Milli * Liter, within=0.03)


def test_deriving_solar_mass() -> None:
    # https://en.wikipedia.org/wiki/Solar_mass#Calculation
    calculated = (4 * π**2 * AstronomicalUnit**3) / (G * JulianYear**2)
    assert calculated == approximately(1 * SolarMass, within=1e-5)


def test_solar_system_masses() -> None:
    assert 1 * EarthMass < 1 * JupiterMass < 1 * SolarMass
    assert 1 * SolarMass == approximately(1047.348644 * JupiterMass, within=3e-4)
    assert 1 * SolarMass == approximately(332946.0487 * EarthMass, within=3e-5)
    assert 1 * JupiterMass == approximately(317.82838 * EarthMass, within=3e-6)


def test_sidereal_day() -> None:
    assert 1 * SiderealDay == approximately(86164.0905 * Second, within=0)
    assert 1 * SiderealDay == approximately(23.9344696 * Hour, within=7e-10)


def test_jansky() -> None:
    assert 1 * Jansky == 1e-26 * Watt / Meter**2 / Hertz


def test_crab() -> None:
    keV = Kilo * ElectronVolt
    cm = Centi * Meter
    s = Second

    assert 1 * Crab == 2.4e-8 * Erg / (cm**2 * s)
    assert 1 * Crab == approximately(15 * keV / (cm**2 * s), within=1e-2)


def test_hubble_constant() -> None:
    assert H0 == approximately(((67400 * Meter) / (1000000 * Parsec)) / Second)


def test_hubble_time() -> None:
    assert 1 * HubbleTime == approximately(14.4e9 * JulianYear, within=0.008)


def test_hubble_length() -> None:
    assert 1 * HubbleLength == approximately(14.4e9 * LightYear, within=0.008)


def test_hubble_volume() -> None:
    assert 1 * HubbleVolume == approximately(1e79 * Meter**3, within=0.09)
    assert 1 * HubbleVolume == approximately(1e31 * LightYear**3, within=0.28)


def test_hubble_barn() -> None:
    assert 1 * HubbleLength * Barn == approximately(13.1 * Liter, within=0.05)



"""
Defines units and constants of use in the field of astronomy, which requires measuring
extremely large distances, masses, and periods of time.

> The [astronomical system of units][1] is a tridimensional system, in that it defines
> units of length, mass and time.  The associated astronomical constants also fix the
> different frames of reference that are needed to report observations.

[1]: https://en.wikipedia.org/wiki/Astronomical_system_of_units

Attributes: Units of length

    AstronomicalUnit (Unit): A unit representing the average distance between the Earth
        and the Sun; it is now defined precisely as 149,597,870,700 meters

    LightYear (Unit): The distance light travels through a vacuum in one `JulianYear`

    Spat (Unit): An obsolete unit of astronomical distance equaling 1 trillion meters

    Parsec (Unit): A measure of large astronomical distances defined in terms of
        [trigonometry and parallax][1]

        [1]: https://en.wikipedia.org/wiki/Parsec

    Siriometer (Unit): An obsolete unit of astronomical distance equaling 1 million AU

Attributes: Units of time

    Day (Unit): Uses the same definition of `Day` as the SI system

    SiderealDay (Unit): The `Time` it takes the Earth to make one rotation with respect
        to distant stars, approximately 86164.0905 seconds, a few minutes shorter than
        the SI Day.

    JulianYear (Unit): defined as 365.25 days of 84,600 SI `Second`

Attributes: Units of mass

    SolarMass (Unit): The mass of the Sun

    EarthMass (Unit): The mass of the Earth

    JupiterMass (Unit): The mass of Jupiter

Attributes: Units for astrophotometry

    Jansky (Unit): A unit of spectral flux density

    Crab (Unit): A unit for measuring the intensity of astromonical X-ray sources, based
        on the intensity of the Crab Nebula

Attributes: Hubble's Law

    H0 (Quantity): Today's estimate of the Hubble Parameter representing the observed
        expansion rate of the universe

    HubbleTime (Unit): An approximation of the age of the universe, if the expansion
        rate had been linear

    HubbleLength (Unit): The distance at which galaxies are receding from us at the
        speed of light

    HubbleVolume (Unit): The sphere within which galaxies are receding from us below
        the speed of light

"""

from measured import Area, Length, Mass, Power, Time, Volume, si
from measured.geometry import π
from measured.physics import c
from measured.si import Kilogram, Mega, Meter, Second, Watt

# Measures of Time

Day = si.Day

# https://en.wikipedia.org/wiki/Sidereal_day
SiderealDay = Time.unit("sidereal day", "SD")
SiderealDay.equals(86164.0905 * Second)

# https://en.wikipedia.org/wiki/Julian_year_(astronomy)
JulianYear = Time.unit("Julian year", "a")
JulianYear.equals(365.25 * Day)


# Measures of Length

AstronomicalUnit = Length.unit("astronomical unit", "AU")
AstronomicalUnit.equals(149597870700 * Meter)

# https://en.wikipedia.org/wiki/Light-year
LightYear = Length.unit("light-year", "ly")
LightYear.equals((c * JulianYear).in_unit(Meter))

# https://en.wikipedia.org/wiki/Spat_(distance_unit)
Spat = Length.unit("spat", "spat")
Spat.equals(1e12 * Meter)

# https://en.wikipedia.org/wiki/Parsec#Calculating_the_value_of_a_parsec
Parsec = Length.unit("parsec", "pc")
Parsec.equals(96939420213600600 / π * Meter)

# https://en.wikipedia.org/wiki/Siriometer
Siriometer = Length.unit("seriometer", "sir")
Siriometer.equals(1000000 * AstronomicalUnit)


# Measures of Mass

# https://en.wikipedia.org/wiki/Solar_mass
SolarMass = Mass.unit("solar mass", "M☉")
SolarMass.equals(1.98847e30 * Kilogram)

EarthMass = Mass.unit("earth mass", "M-earth")
EarthMass.equals(5.9722e24 * Kilogram)

JupiterMass = Mass.unit("jupiter mass", "M-jup")
JupiterMass.equals(1.89813e27 * Kilogram)


# Derived

# https://en.wikipedia.org/wiki/Jansky
Jansky = (Power * Time / Area).unit("jansky", "Jy")
Jansky.equals(1e-26 * Watt * Second / Meter**2)

# https://en.wikipedia.org/wiki/Crab_(unit)
Crab = (Power / Area).unit("crab", "crab")
Crab.equals(2.4e-11 * Watt / Meter**2)


# Hubble's Law

# https://en.wikipedia.org/wiki/Hubble%27s_law#Time-dependence_of_Hubble_parameter
# H0 is today's value of the Hubble parameter
H0 = 67400 * ((Meter / Second) / (Mega * Parsec))

# https://en.wikipedia.org/wiki/Hubble%27s_law#Units_derived_from_the_Hubble_constant
HubbleTime = Time.unit("hubble time", "hubble-time")
HubbleTime.equals(H0**-1)
HubbleTime.equals((H0**-1).in_unit(Second))

HubbleLength = Length.unit("hubble length", "hubble-length")
HubbleLength.equals(c / H0)
HubbleLength.equals(c * HubbleTime)

HubbleVolume = Volume.unit("hubble volume", "hubble-volume")
HubbleVolume.equals(4 / 3 * π * (1 * HubbleLength) ** 3)



"""
Units and non-standardized conventions from the domain of computing.

Attributes: Unit

    Furman (Unit): A unit of `PlaneAngle` that divides the circle into 2¹⁶ divisions,
        making it convenient for representation and arithmetic on a commputer.

    Nibble (Unit): A unit of `Information` equal to half a byte

"""

from measured import Information, PlaneAngle
from measured.geometry import π
from measured.iec import Bit
from measured.si import Radian

# https://en.wikipedia.org/wiki/List_of_unusual_units_of_measurement
# The Furman is a unit of angular measure equal to 1/65,536 of a circle
Furman = PlaneAngle.unit("Furman", "furman")
Furman.equals(1 / 65536 * 2 * π * Radian)

# https://en.wikipedia.org/wiki/Nibble
Nibble = Information.unit("nibble", "nibble")
Nibble.equals(4 * Bit)



import pytest
from hypothesis import assume, given

from measured import One, Quantity, approximately
from measured.hypothesis import quantities


@pytest.fixture(scope="module")
def identity() -> Quantity:
    return 1 * One


@given(a=quantities(), b=quantities(), c=quantities())
def test_associativity(a: Quantity, b: Quantity, c: Quantity) -> None:
    assert (a * b) * c == approximately(a * (b * c))


@given(a=quantities())
def test_identity(identity: Quantity, a: Quantity) -> None:
    assert a * identity == a
    assert identity * a == a


@given(a=quantities())
def test_inverse(identity: Quantity, a: Quantity) -> None:
    assume(a.magnitude != 0)

    inverse = a**-1
    assert inverse * a == a * inverse
    assert inverse * a == approximately(identity)
    assert identity / a == approximately(inverse)


@given(a=quantities(), b=quantities())
def test_commutativity(a: Quantity, b: Quantity) -> None:
    assert a * b == b * a



"""
Constants and functions for computing geometry.

Attributes: Geometrical constants

    π (Quantity): The ratio of a circle's diameter to its radius
"""

import math

from measured import One

π = math.pi * One



from measured import approximately, music
from measured.si import Hertz


def test_middle_A() -> None:
    assert music.A[4] == 440 * Hertz


def test_middle_C() -> None:
    assert music.C[4] == approximately(261.625565 * Hertz)


def test_lowest_c() -> None:
    assert music.C[-1] == approximately(8.175799 * Hertz)


def test_highest_b() -> None:
    assert music.B[10] == approximately(31608.53 * Hertz)


def test_synonyms() -> None:
    assert music.A_sharp == music.B_flat
    assert music.C_sharp == music.D_flat
    assert music.D_sharp == music.E_flat
    assert music.F_sharp == music.G_flat
    assert music.G_sharp == music.A_flat


def test_midi_notes() -> None:
    assert music.MIDI[0] == music.C[-1]
    assert music.MIDI[69] == music.A[4]
    assert music.MIDI[127] == music.G[9]



def test_us_masses_are_based_on_the_avoirdupois_system() -> None:
    from measured import avoirdupois, us

    assert us.Dram is avoirdupois.Dram
    assert us.Grain is avoirdupois.Grain
    assert us.Ounce is avoirdupois.Ounce
    assert us.Pound is avoirdupois.Pound

    assert us.Hundredweight is not avoirdupois.LongHundredweight
    assert (1 * us.Hundredweight) < (1 * avoirdupois.LongHundredweight)

    assert us.Ton is not avoirdupois.LongTon
    assert (1 * us.Ton) < (1 * avoirdupois.LongTon)



import re
import subprocess
from subprocess import CalledProcessError

import pytest


def run(command: str) -> str:
    return subprocess.check_output(command, shell=True, encoding="utf-8")


def test_can_run_as_module() -> None:
    output = run("python -m measured 5 m")
    assert "Magnitude: 5" in output
    assert "Unit: meter" in output


def test_can_run_as_script() -> None:
    output = run("measured 5 m")
    assert "Magnitude: 5" in output
    assert "Unit: meter" in output


def test_prints_help_with_no_arguments() -> None:
    output = run("measured")
    assert "usage: measured [-h]" in output
    assert "positional arguments:" in output


def test_can_list_units() -> None:
    output = run("measured --list")
    assert "m (meter, length)" in output
    assert "in. (inch, length)" in output


def test_can_print_conversions() -> None:
    output = run("measured 1 foot")
    assert "Magnitude: 1" in output
    assert "Unit: foot" in output
    assert re.findall(r"12\.0\s+inch", output)
    assert re.findall(r"0\.3048\s+meter", output)


def test_can_handle_no_conversions() -> None:
    output = run("measured 1 steradian")
    assert "Magnitude: 1" in output
    assert "Unit: steradian" in output


def test_handles_parse_errors_gracefully() -> None:
    with pytest.raises(CalledProcessError) as excinfo:
        run("measured flibbidy gibbidy")

    assert excinfo.value.returncode == 1
    assert "Error parsing quantity" in excinfo.value.stdout



import sys

if sys.version_info < (3, 9):  # pragma: no cover
    # math.gcd changed in Python 3.8 from a two-argument for to a variable argument form
    import math

    from typing_extensions import SupportsIndex

    def recursive_gcd(*integers: SupportsIndex) -> int:
        if len(integers) <= 2:
            return math.gcd(*integers)
        return math.gcd(integers[0], gcd(*integers[1:]))

    _gcd = recursive_gcd

else:  # pragma: no cover
    from math import gcd as _gcd

gcd = _gcd



from measured import One, avoirdupois, troy


def test_grain_is_the_same() -> None:
    assert avoirdupois.Grain == troy.Grain


# https://en.wikipedia.org/wiki/Troy_weight#Troy_ounce_(oz_t)


def test_troy_ounce_is_heavier() -> None:
    assert (1 * troy.Ounce) > 1 * avoirdupois.Ounce


def test_troy_ounce_ratio() -> None:
    assert (1 * troy.Ounce) == 480 / 437.5 * avoirdupois.Ounce


def test_troy_ounce_conversion() -> None:
    (1 * troy.Ounce).in_unit(
        avoirdupois.Ounce
    ) == 1.0971428571428572 * avoirdupois.Ounce

    assert (1 * avoirdupois.Ounce).in_unit(
        troy.Ounce
    ) == 0.9114583333333334 * troy.Ounce


def test_troy_ounce_percentage() -> None:
    assert (1 * troy.Ounce) / (1 * avoirdupois.Ounce) == 1.0971428571428572 * One



from typing import Any, List, Optional, Union

from _pytest.config import Config

from measured import Level, Measurement, Quantity
from measured.conversions import ConversionNotFound

Comparable = Union[Quantity, Level, Measurement]
COMPARABLE = (Quantity, Level, Measurement)


class MeasuredPlugin:
    def pytest_assertrepr_compare(
        self, op: str, left: Any, right: Any
    ) -> Optional[List[str]]:

        if not isinstance(left, COMPARABLE) and not isinstance(right, COMPARABLE):
            return None

        comparison = ""
        difference: Optional[Quantity] = None
        relative = None

        if isinstance(left, COMPARABLE) and isinstance(right, COMPARABLE):
            if isinstance(left, Level):
                left_quantity = left.quantify()
            elif isinstance(left, Measurement):
                left_quantity = left.measurand
            elif isinstance(left, Quantity):
                left_quantity = left

            if isinstance(right, Level):
                right_quantity = right.quantify()
            elif isinstance(right, Measurement):
                right_quantity = right.measurand
            elif isinstance(right, Quantity):
                right_quantity = right

            left_unit = left_quantity.unit

            right_comparison = right

            if isinstance(right, Level):
                try:
                    right_comparison = right_quantity.in_unit(left_unit)
                    difference = left_quantity - right_comparison
                except ConversionNotFound:
                    difference = None

            elif isinstance(right, Measurement):
                try:
                    right_comparison = Measurement(
                        measurand=right_quantity.in_unit(left_unit),
                        uncertainty=right.uncertainty.in_unit(left_unit),
                    )
                    difference = left_quantity - right_comparison.measurand
                except ConversionNotFound:
                    difference = None

            elif isinstance(right, Quantity):
                try:
                    right_comparison = right.in_unit(left_unit)
                    difference = left_quantity - right_comparison
                except ConversionNotFound:
                    difference = None

            comparison = f"{left} {op} {right_comparison}"

            if difference:
                if left_quantity.magnitude != 0:
                    relative = abs(difference.magnitude / left_quantity.magnitude)
                elif right_quantity.magnitude != 0:
                    relative = abs(difference.magnitude / right_quantity.magnitude)
                else:
                    relative = None

        if difference is None:
            difference_description = "difference: (conversion not found)"
        else:
            difference_description = (
                f"difference: absolute={difference} relative={relative})"
            )

        return [
            f"{left} {op} {right}",
            comparison,
            difference_description,
            "where",
            f"{str(left)!r} is {left!r}",
            "and",
            f"{str(right)!r} is {right!r}",
        ]


def pytest_configure(config: Config) -> None:
    config.pluginmanager.register(MeasuredPlugin())



from measured import Length, Mass, Time, approximately
from measured.fff import Firkin, Fortnight, Furlong
from measured.physics import c
from measured.si import (
    Centi,
    Day,
    Hour,
    Kilo,
    Kilogram,
    Mega,
    Meter,
    Micro,
    Minute,
    Second,
)
from measured.us import Inch, Mile, Pound, Yard

# https://en.wikipedia.org/wiki/FFF_system


def test_furlong() -> None:
    assert Furlong.dimension is Length
    assert Furlong.name == "furlong"
    assert Furlong.symbol == "fur."
    assert 1 * Furlong == approximately(201.168 * Meter, 3e-6)
    assert 1 * Furlong == 220 * Yard


def test_firkin() -> None:
    assert Firkin.dimension is Mass
    assert Firkin.name == "firkin"
    assert Firkin.symbol == "fir"
    assert 1 * Firkin == 40.8233133 * Kilogram
    assert 1 * Firkin == 90 * Pound


def test_fortnight() -> None:
    assert Fortnight.dimension is Time
    assert Fortnight.name == "fortnight"
    assert Fortnight.symbol == "ftn"
    assert 1 * Fortnight == 14 * Day
    assert 1 * Fortnight == 1209600 * Second
    assert 1 * Micro * Fortnight == 1.2096 * Second


def test_furlongs_per_fortnight() -> None:
    fpf = 1 * Furlong / Fortnight

    assert fpf == approximately(1.663e-4 * Meter / Second, 6e-5)
    assert fpf == approximately(1 * (Centi * Meter) / Minute, 1 / 400)
    assert fpf == approximately(5.987e-4 * (Kilo * Meter) / Hour, 3e-5)
    assert fpf == approximately((3 / 8) * (Inch / Minute), 5e-2)
    assert fpf == approximately(3.720e-4 * Mile / Hour, 7e-5)

    assert c == approximately(1.8026e12 * Furlong / Fortnight, 8e-6)
    assert c == approximately(1.8026 * (Mega * Furlong) / (Micro * Fortnight), 8e-6)



import cProfile
import functools
import os
from pathlib import Path
from typing import Callable

import typer

from measured import One, Quantity
from measured.astronomical import JulianYear
from measured.si import Ampere, Meter, Ohm, Second, Volt
from measured.us import Ounce, Ton

app = typer.Typer()
options = {"baseline": False}

profiles = Path(".profiles")


@app.callback()
def main(baseline: bool = False) -> None:
    options["baseline"] = baseline


def profiled(func: Callable[[], None]) -> Callable[[], None]:
    @functools.wraps(func)
    def inner() -> None:
        os.makedirs(profiles, exist_ok=True)
        filename = f"{'baseline-' if options['baseline'] else ''}{func.__name__}.prof"
        cProfile.runctx("func()", {"func": func}, {}, str(profiles / filename))

    return inner


@app.command()
@profiled
def quantity_construction() -> None:
    """Tests for constructing Quantities"""
    for a in range(1, 1000001):
        Quantity(a, Meter)


@app.command()
@profiled
def low_equality() -> None:
    """Tests for equality for low-dimensional quantities"""
    for a in (Quantity(a, Meter) for a in range(1, 1001)):
        for b in (Quantity(b, Meter) for b in range(1, 1001)):
            bool(a == b)


@app.command()
@profiled
def high_equality() -> None:
    """Tests for equality for high-dimensional quantities"""
    for a in (Quantity(a, Ohm) for a in range(1, 1001)):
        for b in (Quantity(b, Ohm) for b in range(1, 1001)):
            bool(a == b)


@app.command()
@profiled
def resistances() -> None:
    """Computes a whole lot of Ohms"""
    for a in (Quantity(a, Ampere) for a in range(1, 1001)):
        for v in (Quantity(v, Volt) for v in range(1, 1001)):
            assert v / a == (v.magnitude / a.magnitude) * Ohm


@app.command()
@profiled
def conversions() -> None:
    for o in (Quantity(o, Ounce) for o in range(1, 1001)):
        for t in (Quantity(t, Ton) for t in range(1, 1001)):
            divided = (t / o).in_unit(One)
            assert divided.unit == One
            assert round(divided.magnitude) == round(
                t.magnitude / o.magnitude * 20 * 100 * 16
            )


@app.command()
@profiled
def complex_conversions() -> None:
    for o in (Quantity(o, Ounce / (JulianYear * Ampere)) for o in range(1, 1001)):
        for t in (Quantity(t, Ton / (Second * Ampere)) for t in range(1, 1001)):
            divided = (t / o).in_unit(One)
            assert divided.unit == One


@app.command()
def list_of_quantities() -> None:
    all_the_ohms = []
    for a in (Quantity(a, Ampere) for a in range(1, 10001)):
        for v in (Quantity(v, Volt) for v in range(1, 1001)):
            all_the_ohms.append(v / a)


if __name__ == "__main__":
    app()



import pytest

from measured import (
    Area,
    Length,
    Mass,
    Number,
    Numeric,
    One,
    Pressure,
    Quantity,
    Unit,
    approximately,
    conversions,
)
from measured.geometry import π
from measured.physics import gₙ
from measured.si import Degree, Meter, Minute, Newton, Pascal, Radian, Second
from measured.us import PSI, Acre, Foot, Inch, Pica, Pound, PoundForce, Yard


def test_disallow_self_conversion() -> None:
    with pytest.raises(ValueError, match="unit and itself"):
        Meter.equals(1 * Meter)

    with pytest.raises(ValueError, match="unit and itself"):
        conversions.equate(1 * Meter, 1 * Meter)

    with pytest.raises(ValueError, match="unit and itself"):
        conversions.equate(1 * Meter, 2 * Meter)


def test_disallow_self_zeroing() -> None:
    with pytest.raises(ValueError, match="unit and itself"):
        conversions.translate(Meter, 1 * Meter)


def test_approximating_requires_units_with_conversion() -> None:
    Bogus = Length.unit(name="bogus", symbol="bog")

    one = 1 * Meter
    other = 1 * Bogus

    assert one != other
    assert one != approximately(other)


def test_unit_conversion_must_be_in_same_dimension() -> None:
    with pytest.raises(conversions.ConversionNotFound):
        (1 * Meter).in_unit(Second)


def test_unit_always_converts_to_itself() -> None:
    converted = (10 * Meter).in_unit(Meter)
    assert converted.magnitude == 10
    assert converted.unit == Meter


def test_conversion_with_exponents() -> None:
    assert (1 * Foot) ** 2 == (12 * Inch) ** 2
    assert 1 * Foot**2 == 144 * Inch**2

    assert (1 * Foot) ** 3 == (12 * Inch) ** 3
    assert 1 * Foot**3 == 1728 * Inch**3


def test_conversion_navigates_multiple_steps() -> None:
    # 1 Yard -> 36 Inch -> 216 Pica
    assert 1 * Yard == approximately(216 * Pica)
    assert 216 * Pica == approximately(1 * Yard)


def test_cancelling_units() -> None:
    speed = 10 * Meter / Second
    time = 1 * Minute
    distance = speed * time
    assert distance.in_unit(Meter) == 600 * Meter


def test_dimensionless_units() -> None:
    assert (π / 2 * Radian).in_unit(Degree) == 90 * Degree


def test_conversions_applied_during_subtraction() -> None:
    first = 2 * Meter / Second
    second = 30 * Foot / Minute
    assert first - second == 1.8476 * Meter / Second


@pytest.mark.parametrize(
    "start, end, magnitude",
    [
        (Foot, Inch, 12.0),
        (Foot**2, Meter**2, 0.09290304),
        (Meter**2, Foot**2, 10.76391041670972),
        (Acre, Foot**2, 43560.0),
        (Foot**2, Inch**2, 144.0),
        (Acre, Inch**2, 43560.0 * 144.0),
        (Inch**2, Foot**2, 0.006944444444444444),
        (Foot**2, Acre, 0.00002295684113865932),
        (Inch**2, Acre, 0.006944444444444444 * 0.00002295684113865932),
        (
            Inch**2 * Foot**2,
            Acre**2,
            (0.006944444444444444 * 0.00002295684113865932) / 43560.0,
        ),
    ],
)
def test_conversion_can_navigate_exponents(
    start: Unit, end: Unit, magnitude: Numeric
) -> None:
    converted = conversions.convert(1 * start, end)
    assert converted.unit == end
    assert converted.magnitude == pytest.approx(magnitude)
    assert converted == approximately(magnitude * end)


def test_backtracking_conversions_with_no_path() -> None:
    # Set up some arbitrary units that don't convert to one another, simulating the
    # situation of acres and square inches
    Inchy = Length.unit("inchy", "inchy")
    Achy = Area.unit("achy", "achy")

    with pytest.raises(conversions.ConversionNotFound):
        conversions.convert(1 * Inchy**2, Achy)

    with pytest.raises(conversions.ConversionNotFound):
        conversions.convert(1 * Achy, Inchy**2)


def test_failing_to_convert_numerator() -> None:
    Gud = Length.unit("goodbye", "goodbye")
    Mun = Length.unit("moon", "moon")

    with pytest.raises(conversions.ConversionNotFound):
        (1 * Gud).in_unit(Mun)


def test_replacing_complex_factors_when_there_are_no_alternatives() -> None:
    # This cover a particular branch through conversions._replace_factors, where there
    # are no alternatives for a given unit
    Krikey = Area.unit("krikey", "krikey")
    Mate = Area.unit("mate", "mate")

    with pytest.raises(conversions.ConversionNotFound):
        (1 * Krikey * Krikey).in_unit(Mate * Mate)


def test_alternatives_available_but_not_better() -> None:
    # This cover a particular branch through conversions._replace_factors, where there
    # are multiple alternatives, but none is particularly better
    Oy = Area.unit("oy", "oy")
    Vey = Area.unit("vey", "vey")
    Oof = Area.unit("oof", "oof")

    Oy.equals(1 * Oof)
    Vey.equals(1 * Oof)

    assert 1 * Oof * Second == 1 * Oy * Second


def test_failing_to_convert_denominator() -> None:
    Flib = Mass.unit("flibbidy", "flibbidy")
    Flob = Mass.unit("flobbidy", "flobbidy")

    assert (Meter / Flib).dimension is (Meter / Flob).dimension

    with pytest.raises(conversions.ConversionNotFound):
        (1 * Meter / Flib).in_unit(Meter / Flob)


def test_failing_to_collapse_dimensions() -> None:
    Jib = Mass.unit("jibbity", "jibbity")
    Job = Mass.unit("jobbity", "jobbity")

    assert (Jib / Job).dimension is Number
    assert (Job / Jib).dimension is Number

    with pytest.raises(conversions.ConversionNotFound):
        ((1 * Jib) / (1 * Job)).in_unit(One)


def test_replacing_multiple_factors() -> None:
    left = 1 * PoundForce * PoundForce
    right = (1 * Pound * gₙ) * (1 * Pound * gₙ)

    # These comparisons require replacing the PoundForce unit multiple times
    assert left == right
    assert right == left


PRESSURE_EQUIVALENTS = [
    10000 * Newton / Meter**2,
    10000 * Pascal,
    1.4503774 * PSI,
    1.4503774 * PoundForce / Inch**2,
    208.85434305 * PoundForce / Foot**2,
    22046.226 * (Pound * Meter / Second**2) / Meter**2,
    2048.163 * (Pound * Meter / Second**2) / Foot**2,
    6719.689751 * (Pound * Foot / Second**2) / Foot**2,
    46.664512 * (Pound * Foot / Second**2) / Inch**2,
    167992.2432 * (Pound * Foot / Minute**2) / Inch**2,
    9097695 * PoundForce / Acre,
    89217910.67175 * (Pound * Meter / Second**2) / Acre,
]


@pytest.mark.parametrize("equivalent", PRESSURE_EQUIVALENTS)
def test_converting_pressure_terms_forward(equivalent: Quantity) -> None:
    quantity = 10000 * Newton / Meter**2
    assert quantity.unit.dimension is Pressure

    converted = quantity.in_unit(equivalent.unit)
    assert converted.unit.dimension is Pressure
    assert converted.unit is equivalent.unit
    assert converted.magnitude == pytest.approx(equivalent.magnitude)


@pytest.mark.parametrize("equivalent", PRESSURE_EQUIVALENTS)
def test_converting_pressure_terms_backward(equivalent: Quantity) -> None:
    quantity = 10000 * Newton / Meter**2
    assert quantity.unit.dimension is Pressure

    reversed = equivalent.in_unit(Newton / Meter**2)
    assert reversed.unit.dimension is Pressure
    assert reversed.unit is Newton / Meter**2
    assert reversed.magnitude == pytest.approx(10000)
